#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Run a test of the entire code on dummy data. """
import numpy as np
import h5py
import os
import shutil
from unittest import TestCase

from orcanet.core import Organizer
from orcanet.model_builder import ModelBuilder
from orcanet_contrib.custom_objects import get_custom_objects


class TestIntegration(TestCase):
    """ Run the actual training to see if it throws any errors. """
    def setUp(self):
        """
        Make a .temp directory in the current working directory, generate
        dummy data in it and set up the cfg object.

        """
        # Pathes
        # Temporary output folder
        self.temp_dir = os.path.join(os.path.dirname(__file__),
                                     ".temp", "full_test")
        self.output_folder = os.path.join(self.temp_dir, "model")

        # Pathes to temp dummy data that will get generated
        train_inp = (self.temp_dir + "/train1.h5", self.temp_dir + "/train2.h5")
        self.train_pathes = {"testing_input": train_inp}

        val_inp = (self.temp_dir + "/val1.h5", self.temp_dir + "/val2.h5")
        self.val_pathes = {"testing_input": val_inp}

        # The config file to load
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        config_file = os.path.join(self.data_folder, "config_test.toml")

        # Make sure temp dir does either not exist or is empty
        if os.path.exists(self.temp_dir):
            assert len(os.listdir(self.temp_dir)) == 0
        else:
            os.makedirs(self.temp_dir)

        # Make dummy data of given shape
        self.shape = (3, 3, 3, 3)
        for path1, path2 in (train_inp, val_inp):
            make_dummy_data(path1, path2, self.shape)

        def make_orga():
            orga = Organizer(self.output_folder, config_file=config_file)
            orga.cfg._train_files = self.train_pathes
            orga.cfg._val_files = self.val_pathes
            orga.cfg._list_file = "test.toml"
            orga.cfg.zero_center_folder = self.temp_dir
            orga.cfg.label_modifier = label_modifier
            orga.cfg.custom_objects = get_custom_objects()
            return orga

        self.make_orga = make_orga

    def tearDown(self):
        """ Remove the .temp directory. """
        shutil.rmtree(self.temp_dir)

    def test_integration_zero_center(self):
        """ Calculate the zero center image and check if it works properly. """
        orga = self.make_orga()
        xs_mean = orga.get_xs_mean()
        target_xs_mean = np.ones(self.shape)/4
        self.assertTrue(np.allclose(xs_mean["testing_input"], target_xs_mean))

        file = orga.cfg.zero_center_folder + "/" + orga.cfg._list_file + \
            '_input_' + "testing_input" + '.npz'
        zero_center_used_ip_files = np.load(file)['zero_center_used_ip_files']
        self.assertTrue(np.array_equal(zero_center_used_ip_files,
                                       orga.cfg._train_files["testing_input"]))

    def test_integration_multi_input_model(self):
        """
        Run whole script on dummy data to see if it throws an error.

        Build a model with the ModelBuilder.
        Train for 2 epochs.
        Reset organizer.
        Resume for 1 epoch with different lr and sample modifier.
        Predict.

        """
        orga = self.make_orga()

        model_file = os.path.join(self.data_folder, "model_test.toml")
        builder = ModelBuilder(model_file)
        initial_model = builder.build(orga)

        orga.train_and_validate(initial_model, epochs=2)

        def test_learning_rate(epoch, fileno):
            lr = 0.001 * (epoch + 0.1*fileno)
            return lr

        def test_modifier(xs):
            xs = {key: xs[key] for key in xs}
            return xs

        orga = self.make_orga()
        orga.cfg.learning_rate = test_learning_rate
        orga.cfg.sample_modifier = test_modifier
        orga.train_and_validate(epochs=1)
        orga.predict()


def make_dummy_data(filepath1, filepath2, shape):
    """
    Make a total of 100 ones vs 300 zeroes of dummy data over two files.

    Parameters
    ----------
    filepath1 : str
        Path to file 1.
    filepath2 : str
        Path to file 2.
    shape : tuple
        Shape of the data, not including sample dimension.

    """
    xs1 = np.concatenate([np.ones((75,) + shape), np.zeros((75,) + shape)])
    xs2 = np.concatenate([np.ones((25,) + shape), np.zeros((225,) + shape)])

    dtypes = [('event_id', '<f8'), ('particle_type', '<f8'), ('energy', '<f8'),
              ('is_cc', '<f8'), ('bjorkeny', '<f8'), ('dir_x', '<f8'),
              ('dir_y', '<f8'), ('dir_z', '<f8'), ('time_interaction', '<f8'),
              ('run_id', '<f8'), ('vertex_pos_x', '<f8'), ('vertex_pos_y', '<f8'),
              ('vertex_pos_z', '<f8'), ('time_residual_vertex', '<f8'),
              ('prod_ident', '<f8'), ('group_id', '<i8')]
    ys1 = np.ones((150, 16)).ravel().view(dtype=dtypes)
    ys2 = np.ones((250, 16)).ravel().view(dtype=dtypes)

    for xs, ys, filepath in ((xs1, ys1, filepath1), (xs2, ys2, filepath2)):
        h5f = h5py.File(filepath, 'w')
        h5f.create_dataset('x', data=xs, dtype='uint8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)
        h5f.close()


def label_modifier(y_values):
    ys = dict()

    ys['dx'], ys['dx_err'] = y_values['dir_x'], y_values['dir_x']

    for key_label in ys:
        ys[key_label] = ys[key_label].astype(np.float32)
    return ys



"""
Use orga.train with a parser.

Usage:
    parser_orcatrain.py [options] FOLDER LIST CONFIG MODEL
    parser_orcatrain.py (-h | --help)

Arguments:
    FOLDER  Path to the folder where everything gets saved to, e.g. the
            summary.txt, the plots, the trained models, etc.
    LIST    A .toml file which contains the pathes of the training and
            validation files. An example can be found in
            examples/example_list.toml
    CONFIG  A .toml file which sets up the training. An example can be
            found in examples/example_config.toml. The possible parameters
            are listed in core.py in the class Configuration.
    MODEL   Path to a .toml file with infos about a model.
            An example can be found in examples/example_model.toml.

Options:
    -h --help    Show this screen.
    --recompile  Recompile the keras model, e.g. needed if the loss weights
                 are changed during the training.

"""
from docopt import docopt
import keras as ks

from orcanet.core import Organizer
from model_builder import ModelBuilder
from orcanet_contrib.orca_handler_util import orca_learning_rates, update_objects


def orca_train(output_folder, list_file, config_file, model_file,
               recompile_model=False):
    """
    Run orga.train with predefined ModelBuilder networks using a parser.

    Parameters
    ----------
    output_folder : str
        Path to the folder where everything gets saved to, e.g. the summary
        log file, the plots, the trained models, etc.
    list_file : str
        Path to a list file which contains pathes to all the h5 files that
        should be used for training and validation.
    config_file : str
        Path to a .toml file which overwrites some of the default settings
        for training and validating a model.
    model_file : str
        Path to a file with parameters to build a model of a predefined
        architecture with OrcaNet.
    recompile_model : bool
        If the model should be recompiled or not. Necessary, if e.g. the
        loss_weights are changed during the training.

    """
    # Set up the Organizer with the input data
    orga = Organizer(output_folder, list_file, config_file)

    # Load in the orga sample-, label-, and dataset-modifiers, as well as
    # the custom objects
    update_objects(orga, model_file)

    # If this is the start of the training, a compiled model needs to be
    # handed to the orga.train function
    if orga.io.get_latest_epoch() is None:
        # The ModelBuilder class allows to construct models from a toml file,
        # adapted to the datasets in the orga instance. Its modifiers will
        # be taken into account for this
        builder = ModelBuilder(model_file)
        model = builder.build(orga)

    elif recompile_model is True:
        builder = ModelBuilder(model_file)

        path_of_model = orga.io.get_model_path(-1, -1)
        model = ks.models.load_model(path_of_model,
                                     custom_objects=orga.cfg.custom_objects)
        print("Recompiling the saved model")
        model = builder.compile_model(model)

    else:
        model = None

    # Use a custom LR schedule
    orga.cfg.learning_rate = orca_learning_rates("triple_decay",
                                                 orga.io.get_no_of_files("train"))

    # start the training
    orga.train_and_validate(model=model)


def parse_input():
    """ Run the orca_train function with a parser. """
    args = docopt(__doc__)
    output_folder = args['FOLDER']
    list_file = args['LIST']
    config_file = args['CONFIG']
    model_file = args['MODEL']
    recompile_model = args['--recompile']
    orca_train(output_folder, list_file, config_file, model_file,
               recompile_model=recompile_model)


if __name__ == '__main__':
    parse_input()



from unittest import TestCase
from unittest.mock import MagicMock
import os
import h5py
import numpy as np
from keras.models import Model
from keras.layers import Dense, Input, Concatenate, Flatten
from keras.callbacks import LambdaCallback

from orcanet.core import Organizer
from orcanet.backend import hdf5_batch_generator, get_datasets, get_learning_rate, train_model, validate_model, make_model_prediction
from orcanet.utilities.nn_utilities import get_auto_label_modifier


class TestBatchGenerator(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = os.path.join(os.path.dirname(__file__), ".temp",
                                    "test_backend")
        os.mkdir(cls.temp_dir)

        # make some dummy data
        cls.n_bins = {'input_A': (2, 3), 'input_B': (2, 3)}
        cls.train_sizes = [3, 5]
        cls.train_A_file_1 = {
            "path": cls.temp_dir + "/input_A_train_1.h5",
            "shape": cls.n_bins["input_A"],
            "size": cls.train_sizes[0],
        }

        cls.train_B_file_1 = {
            "path": cls.temp_dir + "/input_B_train_1.h5",
            "shape": cls.n_bins["input_B"],
            "size": cls.train_sizes[0],
        }

        cls.filepaths_file_1 = {
            "input_A": cls.train_A_file_1["path"],
            "input_B": cls.train_B_file_1["path"],
        }

        cls.train_A_file_1_ctnt = save_dummy_h5py(**cls.train_A_file_1)
        cls.train_B_file_1_ctnt = save_dummy_h5py(**cls.train_B_file_1)

    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.output_folder = self.data_folder + "/dummy_model"

        list_file = self.data_folder + "/in_out_test_list.toml"
        config_file = None

        self.orga = Organizer(self.output_folder, list_file, config_file)
        self.batchsize = 2
        self.orga.cfg.batchsize = self.batchsize
        self.orga.cfg.label_modifier = label_modifier

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.train_A_file_1["path"])
        os.remove(cls.train_B_file_1["path"])
        os.rmdir(cls.temp_dir)

    def test_batch(self):
        filepaths = self.filepaths_file_1
        gene = hdf5_batch_generator(self.orga, filepaths)

        target_xs_batch_1 = {
            "input_A": self.train_A_file_1_ctnt[0][:2],
            "input_B": self.train_B_file_1_ctnt[0][:2],
        }

        target_ys_batch_1 = label_modifier(self.train_A_file_1_ctnt[1][:2])

        target_xs_batch_2 = {
            "input_A": self.train_A_file_1_ctnt[0][2:],
            "input_B": self.train_B_file_1_ctnt[0][2:],
        }

        target_ys_batch_2 = label_modifier(self.train_A_file_1_ctnt[1][2:])

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

        # appended batches
        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

        with self.assertRaises(StopIteration):
            next(gene)

    def test_batch_zero_center(self):
        filepaths = self.filepaths_file_1

        xs_mean = {name: np.ones(shape) * 0.5 for name, shape in self.n_bins.items()}

        self.orga.get_xs_mean = MagicMock(return_value=xs_mean)
        gene = hdf5_batch_generator(self.orga, filepaths, zero_center=True)

        target_xs_batch_1 = {
            "input_A": np.subtract(self.train_A_file_1_ctnt[0][:2], xs_mean["input_A"]),
            "input_B": np.subtract(self.train_B_file_1_ctnt[0][:2], xs_mean["input_B"]),
        }

        target_ys_batch_1 = label_modifier(self.train_A_file_1_ctnt[1][:2])

        target_xs_batch_2 = {
            "input_A": np.subtract(self.train_A_file_1_ctnt[0][2:], xs_mean["input_A"]),
            "input_B": np.subtract(self.train_B_file_1_ctnt[0][2:], xs_mean["input_B"]),
        }

        target_ys_batch_2 = label_modifier(self.train_A_file_1_ctnt[1][2:])

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

    def test_batch_sample_modifier(self):
        filepaths = self.filepaths_file_1

        def sample_modifier(xs_in):
            mod = {name: val*2 for name, val in xs_in.items()}
            return mod

        self.orga.cfg.sample_modifier = sample_modifier
        gene = hdf5_batch_generator(self.orga, filepaths)

        target_xs_batch_1 = {
            "input_A": self.train_A_file_1_ctnt[0][:2]*2,
            "input_B": self.train_B_file_1_ctnt[0][:2]*2,
        }

        target_ys_batch_1 = label_modifier(self.train_A_file_1_ctnt[1][:2])

        target_xs_batch_2 = {
            "input_A": self.train_A_file_1_ctnt[0][2:]*2,
            "input_B": self.train_B_file_1_ctnt[0][2:]*2,
        }

        target_ys_batch_2 = label_modifier(self.train_A_file_1_ctnt[1][2:])

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)

        xs, ys = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)

    def test_batch_mc_infos(self):
        filepaths = self.filepaths_file_1

        gene = hdf5_batch_generator(self.orga, filepaths, yield_mc_info=True)

        target_xs_batch_1 = {
            "input_A": self.train_A_file_1_ctnt[0][:2],
            "input_B": self.train_B_file_1_ctnt[0][:2],
        }

        target_ys_batch_1 = label_modifier(self.train_A_file_1_ctnt[1][:2])
        target_mc_info_batch_1 = self.train_A_file_1_ctnt[1][:2]

        target_xs_batch_2 = {
            "input_A": self.train_A_file_1_ctnt[0][2:],
            "input_B": self.train_B_file_1_ctnt[0][2:],
        }

        target_ys_batch_2 = label_modifier(self.train_A_file_1_ctnt[1][2:])
        target_mc_info_batch_2 = self.train_A_file_1_ctnt[1][2:]

        xs, ys, mc_info = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_1)
        assert_dict_arrays_equal(ys, target_ys_batch_1)
        assert_equal_struc_array(mc_info, target_mc_info_batch_1)

        xs, ys, mc_info = next(gene)
        assert_dict_arrays_equal(xs, target_xs_batch_2)
        assert_dict_arrays_equal(ys, target_ys_batch_2)
        assert_equal_struc_array(mc_info, target_mc_info_batch_2)


class TestFunctions(TestCase):
    def test_get_datasets(self):
        mc_info = "mc_info gets simply passed forward"
        y_true = {
            "out_A": 1,
            "out_B": 2,
        }
        y_pred = {
            "out_pred_A": 3,
            "out_pred_B": 4,
        }
        target = {
            "mc_info": mc_info,
            "label_out_A": 1,
            "label_out_B": 2,
            "pred_out_pred_A": 3,
            "pred_out_pred_B": 4,
        }
        datasets = get_datasets(mc_info, y_true, y_pred)
        self.assertDictEqual(datasets, target)

    def test_get_learning_rate_float(self):
        user_lr = 0.1
        no_train_files = 3

        for fileno in range(no_train_files):
            lr = get_learning_rate((1, fileno), user_lr, no_train_files)
            self.assertEqual(lr, user_lr)

        for fileno in range(no_train_files):
            lr = get_learning_rate((6, fileno), user_lr, no_train_files)
            self.assertEqual(lr, user_lr)

    def test_get_learning_rate_tuple(self):
        user_lr = (0.1, 0.2)
        no_train_files = 3

        rates_epoch_1 = [0.1, 0.08, 0.064]
        for fileno in range(no_train_files):
            lr = get_learning_rate((1, fileno+1), user_lr, no_train_files)
            self.assertAlmostEqual(lr, rates_epoch_1[fileno])

        rates_epoch_2 = [0.0512, 0.04096, 0.032768]
        for fileno in range(no_train_files):
            lr = get_learning_rate((2, fileno + 1), user_lr, no_train_files)
            self.assertAlmostEqual(lr, rates_epoch_2[fileno])

    def test_get_learning_rate_function(self):
        def get_lr(epoch, filenos):
            return epoch+filenos

        user_lr = get_lr
        no_train_files = 3

        rates_epoch_1 = [2, 3, 4]
        for fileno in range(no_train_files):
            lr = get_learning_rate((1, fileno+1), user_lr, no_train_files)
            self.assertAlmostEqual(lr, rates_epoch_1[fileno])

        rates_epoch_2 = [3, 4, 5]
        for fileno in range(no_train_files):
            lr = get_learning_rate((2, fileno + 1), user_lr, no_train_files)
            self.assertAlmostEqual(lr, rates_epoch_2[fileno])

    def test_get_learning_rate_other(self):
        user_lr = print
        no_train_files = 3

        with self.assertRaises(TypeError):
            get_learning_rate((1, 1), user_lr, no_train_files)


class TestTrainValidatePredict(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = os.path.join(os.path.dirname(__file__), ".temp",
                                    "test_backend")
        os.mkdir(cls.temp_dir)
        cls.file_sizes = [500, ]
        # make some dummy data
        cls.inp_A_file = {
            "path": cls.temp_dir + "/input_A_file.h5",
            "shape": (2, 3),
            "size": cls.file_sizes[0],
        }

        cls.inp_B_file = {
            "path": cls.temp_dir + "/input_B_file.h5",
            "shape": (3, 4),
            "size": cls.file_sizes[0],
        }

        cls.filepaths = {
            "input_A": (cls.inp_A_file["path"],),
            "input_B": (cls.inp_B_file["path"],),
        }

        cls.input_shapes = {
            "input_A": cls.inp_A_file["shape"],
            "input_B": cls.inp_B_file["shape"],
        }

        cls.output_shapes = {
            "mc_A": 1,
            "mc_B": 1,
        }

        cls.train_A_file_1_ctnt = save_dummy_h5py(**cls.inp_A_file, mode="half")
        cls.train_B_file_1_ctnt = save_dummy_h5py(**cls.inp_B_file, mode="half")

    def setUp(self):
        self.orga = Organizer("./")
        self.orga.cfg.batchsize = 9

        self.orga.io.get_local_files = MagicMock(return_value=self.filepaths)
        self.orga.io.get_file_sizes = MagicMock(return_value=self.file_sizes)

        self.model = build_dummy_model(self.input_shapes, self.output_shapes)
        self.model.compile(loss="mse", optimizer="sgd")
        self.orga._auto_label_modifier = get_auto_label_modifier(self.model)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.inp_A_file["path"])
        os.remove(cls.inp_B_file["path"])
        os.rmdir(cls.temp_dir)

    def test_train(self):
        epoch = (1, 1)
        batch_nos = []
        batch_print_callback = LambdaCallback(
            on_batch_begin=lambda batch, logs: batch_nos.append(batch))
        self.orga.cfg.callback_train = batch_print_callback

        history = train_model(self.orga, self.model, epoch, batch_logger=False)
        target = {
            'loss': 18.105026489263054,
            'mc_A_loss': 9.569378078221458,
            'mc_B_loss': 8.535648400507155,
        }
        self.assertDictEqual(history, target)
        self.assertSequenceEqual(batch_nos, list(range(int(self.file_sizes[0]/self.orga.cfg.batchsize))))

    def test_validate(self):
        history = validate_model(self.orga, self.model)
        # input to model is ones
        # --> after concatenate: 18 ones
        # Output of each output layer = 18
        # labels: 0 and 1
        # --> loss (18-0)^2 and (18-1)^2
        target = {
            'loss': 18**2 + 17**2,
            'mc_A_loss': 18**2,
            'mc_B_loss': 17**2,
        }
        self.model.summary()
        self.assertDictEqual(history, target)

    def test_predict(self):
        eval_filename = self.temp_dir + "/pred.h5"
        try:
            make_model_prediction(self.orga, self.model, eval_filename)
            file_cntn = {}
            with h5py.File(eval_filename) as file:
                for key in file.keys():
                    file_cntn[key] = np.array(file[key])
        finally:
            os.remove(eval_filename)

        target_datasets = [
            'label_mc_A', 'label_mc_B', 'mc_info', 'pred_mc_A', 'pred_mc_B'
        ]
        target_shapes = [
            (500,), (500,), (500,), (500, 1), (500, 1),
        ]
        target_contents = [
            np.zeros(target_shapes[0]),
            np.ones(target_shapes[1]),
            self.train_A_file_1_ctnt[1],
            np.ones(target_shapes[3]) * 18,
            np.ones(target_shapes[4]) * 18,
        ]
        target_mc_names = ('mc_A', 'mc_B')

        datasets = list(file_cntn.keys())
        shapes = [file_cntn[key].shape for key in datasets]
        mc_dtype_names = file_cntn["mc_info"].dtype.names

        self.assertSequenceEqual(datasets, target_datasets)
        self.assertSequenceEqual(shapes, target_shapes)
        self.assertSequenceEqual(mc_dtype_names, target_mc_names)
        for i, key in enumerate(target_datasets):
            value = file_cntn[key]
            target = target_contents[i]
            np.testing.assert_array_equal(value, target)


def save_dummy_h5py(path, shape, size, mode="asc"):
    """
    if mode == asc:
        xs[0] is np.zeros, xs[1] is np.ones, etc
        ys[0] is ndarray with 0.5s, ys[1] with 1.5s etc

    if mode == half:
        xs is ones,
        ys[:, 0] is 0, ys[:, 1] is 1

    """
    xs = np.ones((size,) + shape)

    dtypes = [('mc_A', '<f8'), ('mc_B', '<f8'), ]
    ys = np.ones((size, 2))

    if mode == "asc":
        for sample_no in range(len(xs)):
            xs[sample_no, ...] = sample_no
            ys[sample_no, ...] = sample_no + 0.5

    elif mode == "half":
        ys[:, 0] = 0.

    else:
        raise AssertionError

    ys = ys.ravel().view(dtype=dtypes)

    with h5py.File(path, 'w') as h5f:
        h5f.create_dataset('x', data=xs, dtype='<f8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)
    return xs, ys


def label_modifier(y_values):
    ys = dict()
    for name in y_values.dtype.names:
        ys[name] = y_values[name]
    return ys


def assert_dict_arrays_equal(a, b):
    for key, value in a.items():
        assert key in b
        np.testing.assert_array_almost_equal(a[key], b[key])


def assert_equal_struc_array(a, b):
    """  np.testing.assert_array_equal does not work for arrays containing nans...
     so test individual instead. """
    np.testing.assert_array_equal(a.dtype, b.dtype)
    for name in a.dtype.names:
        np.testing.assert_almost_equal(a[name], b[name])


def build_dummy_model(input_shapes, output_shapes):
    """

    Parameters
    ----------
    input_shapes : dict
    output_shapes : dict

    Returns
    -------
    model : keras model

    """
    inputs, flattend = [], []
    for name, shape in input_shapes.items():
        inp = Input(shape, name=name)
        flat = Flatten()(inp)
        inputs.append(inp)
        flattend.append(flat)

    conc = Concatenate()(flattend)

    outputs = []
    for name, shape in output_shapes.items():
        outputs.append(Dense(shape, name=name, kernel_initializer="Ones",
                             bias_initializer="Zeros")(conc))

    model = Model(inputs, outputs)
    return model



# -*- coding: utf-8 -*-
"""
Multi-gpu code for Keras/TF.
From https://github.com/avolkov1/keras_experiments
"""
# MODIFIED. Inspiration taken from the ref link below.
# ref: https://raw.githubusercontent.com/kuza55/keras-extras/master/utils/multi_gpu.py @IgnorePep8
# The inspirational one carried license:
#     Apache License
#     Version 2.0, January 2004
# For further info refer to: https://github.com/kuza55/keras-extras
#
# Also used https://github.com/fchollet/keras/issues/2436 which was just
# posted as code snippets in a forum.


import sys
# import time
from itertools import chain

import warnings

from .multi_gpu_utils import Capturing

from keras import backend as KB
from keras.layers.core import Lambda
from keras.models import Model
from keras.layers.merge import Concatenate  # , Average)
# import keras.layers as KL
import keras.optimizers as KO

if KB.backend() == 'tensorflow':
    # Monkey patch Keras back-end to use Function with enqueue.
    # import keras_exp._patch_tf_backend as tfbpatch
    # tfbpatch.patch()
    # from keras_exp._patch_tf_backend import patch as tfbpatch
    # tfbpatch()

    import tensorflow as tf
    from tensorflow.python.client import device_lib

    try:
        from tensorflow.contrib import nccl
        have_nccl = True
        print('NCCL support available', file=sys.stderr)
    except ImportError:
        have_nccl = False
        print('WARNING: NCCL support not available', file=sys.stderr)

    from tensorflow.python.ops import data_flow_ops


_DEBUG = False

__all__ = ('get_available_gpus', 'make_parallel', 'print_mgpu_modelsummary',
           'ModelMGPU')


def get_available_gpus(ngpus=-1):
    '''
    :param int ngpus: GPUs max to use. Default -1 means all gpus.
    :returns: List of gpu devices. Ex.: ['/gpu:0', '/gpu:1', ...]
    '''
    local_device_protos = device_lib.list_local_devices()
    gpus_list = [x.name for x in local_device_protos if x.device_type == 'GPU']
    return gpus_list[:ngpus] if ngpus > -1 else gpus_list


def print_mgpu_modelsummary(model):
    '''Prints the summary for a multi-GPU keras model.
    :param model: Keras model.
    :type model: Model
    '''
    # print json.dumps(model.get_config(), indent=2)  # DEBUG
    print('\nMULTI-GPU MODEL: {}'.format(model.name))
    print(model.summary())
    for layer in model.layers:
        # print 'layer:', layer, '\ttype:', type(layer)
        if isinstance(layer, Model):
            submodel = layer
            print('\n\tSUBMODEL SUMMARY: {}'.format(layer.name))
            with Capturing() as msum:
                minfo = submodel.summary()
            print('\t{}\n\t{}\n'.format('\n\t'.join(msum), minfo))


def all_sync_params(tower_params, devices, usenccl=True):
    """Assigns the params from the first tower to all others"""
    if len(devices) == 1:
        return tf.no_op()
    sync_ops = []
    if have_nccl and usenccl:
        for param_on_devices in zip(*tower_params):
            # print('PARAM_ON_DEVICES: {}'.format(param_on_devices))  # DEBUG
            # Note: param_on_devices is [paramX_gpu0, paramX_gpu1, ...]
            param0 = param_on_devices[0]
            send_op, received_tensors = nccl.broadcast(param0, devices[1:])
            sync_ops.append(send_op)
            for device, param, received in zip(devices[1:],
                                               param_on_devices[1:],
                                               received_tensors):
                with tf.device(device):
                    sync_op = param.assign(received)
                    sync_ops.append(sync_op)
    else:
        params0 = tower_params[0]
        for device, params in zip(devices, tower_params):
            with tf.device(device):
                for param, param0 in zip(params, params0):
                    sync_op = param.assign(param0.read_value())
                    sync_ops.append(sync_op)

    return tf.group(*sync_ops)


# def stage(tensors):
#     """Stages the given tensors in a StagingArea for asynchronous put/get.
#     """
#     stage_area = data_flow_ops.StagingArea(
#         dtypes=[tensor.dtype for tensor in tensors],
#         shapes=[tensor.get_shape() for tensor in tensors])
#     put_op = stage_area.put(tensors)
#     get_tensors = stage_area.get()
#     if not isinstance(get_tensors, list):
#         get_tensors = [get_tensors]
#     # print('GET_TENSORS: {}'.format(get_tensors))  # DEBUG
#
#     get_tensors = [tf.reshape(gt, t.get_shape())
#                    for (gt, t) in zip(get_tensors, tensors)]
#     return put_op, get_tensors


class ModelMGPU(Model):
    '''Override load and save methods of the multi-gpu model. The load and
    save should correspond to the serial model's load and save.
    If there are other idiosyncracies to handle for multi-gpu model case then
    these can be handled in this subclass. A serial model should always be
    instantiated prior to wrapping it or converting it to a multi-GPU model.
    This multi-gpu implementation uses data-parallelism.
    A copy-constructor is not implemented so optionally pass any additional
    parameters besides inputs/outputs as args/kwargs to initialize the
    multi-gpu model the same way as the serial model. Typically not needed.
    Currently, it seems that using NCCL and synchronizing/averaging gradients
    slows multi-gpu processing down.
    .. seealso::
        Refer to :func:`make_parallel` docstring for scenarios when
        out-of-memory errors might occur and workaround.
    Kwargs:
    :param Model serial_model: Serial i.e. non-multi GPU Keras model. REQUIRED.
    :param list gdev_list: List of gpu devices i.e. ['/gpu:0', '/gpu:1', ...]
        Use function get_available_gpus to get the list of available gpus.
        This can be a list of strings or list of instances of tf.DeviceSpec.
        REQUIRED.
    :param str ps_device: Parameter server device to use.
    :param bool usenccl: Use the contrib.nccl Tensorflow library for initial
        parameter synchronization and gradients averaging. Note, the models
        usenccl option overrides the optimizers usenccl option.
        Default: False
        Raises RuntimeError if specified True and a non-multi-gpu optimizer is
        passed during compile stage.
    :param bool initsync: Synchronize initial Variables i.e. weights,
        biases, etc. Default: True
    :param bool syncopt: Synchronize gradients. Requires a multi-gpu optimizer.
        Default: False
    :param bool enqueue: Use StagingArea in the multi-GPU model. Could
        potentially speed up Host-to-Device transfers.
        Produces a warning that kwargs are ignored for Tensorflow. The
        _patch_tf_backend module mokey patches the Function in
        tensorflow_backend to use the enqueue_ops option.
        Default: False
    '''
    def __init__(self, *args, **kwargs):
        # :param model_creator: Callable that returns a serial i.e. non-multi
        #     GPU Keras model i.e. a keras.models.Model model. REQUIRED.
        #     Suggestion, use partial from functools to setup model_creator.
        # try:
        #     model_creator = kwargs.pop('model_creator')
        # except KeyError:
        #     raise RuntimeError('Keyword argument "model_creator" required '
        #                        'for ModelMGPU.')

        try:
            smodel = kwargs.pop('serial_model')
        except KeyError:
            raise RuntimeError('Keyword argument "serial_model" required '
                               'for ModelMGPU.')

        # SET STATE: Instance of serial model for checkpointing
        self._smodel = smodel  # model_creator()

        try:
            gdev_list = kwargs.pop('gdev_list')
        except KeyError:
            raise RuntimeError('Keyword argument "gdev_list" required '
                               'for ModelMGPU.')
        self._gdev_list = gdev_list

        mname = kwargs.pop('name', self._smodel.name)
        kwargs['name'] = mname

        self._ps_device = kwargs.pop('ps_device', '/cpu:0')
        self._initsync = kwargs.pop('initsync', True)
        self._usenccl = kwargs.pop('usenccl', False)
        self._syncopt = kwargs.pop('syncopt', False)
        self._enqueue = kwargs.pop('enqueue', False)

        if self._enqueue:
            warnings.warn('Enqueue option to use StagingArea currenctly does '
                          'not work.', UserWarning)

        # NOTE: To use staging have to patch keras tensorflow_backend.Function.
        #     Function implementation in keras_exp.multigpu._patch_tf_backend
        self._enqueue_ops = []

        self._tower_params = []  # For init/sync'ing of parameters.
        self._init_make_dataparallel(gdev_list, *args,
                                     **kwargs)

    def __getattribute__(self, attrname):
        '''Override load and save methods to be used from the serial-model. The
        serial-model holds references to the weights in the multi-gpu model.
        '''
        # return Model.__getattribute__(self, attrname)
        if 'load' in attrname or 'save' in attrname:
            return getattr(self._smodel, attrname)

        return super(ModelMGPU, self).__getattribute__(attrname)

    # ref: https://github.com/fchollet/keras/issues/2436
    def _init_make_dataparallel(self, gdev_list, *args, **kwargs):
        '''Uses data-parallelism to convert a serial model to multi-gpu. Refer
        to make_parallel doc.
        '''
        gpucopy_ops = []

        def slice_batch(x, ngpus, part, dev):
            '''Divide the input batch into [ngpus] slices, and obtain slice
            no. [part]. i.e. if len(x)=10, then slice_batch(x, 2, 1) will
            return x[5:].
            '''
            sh = KB.shape(x)
            L = sh[0] // ngpus
            if part == ngpus - 1:
                xslice = x[part * L:]
            else:
                xslice = x[part * L:(part + 1) * L]

            # tf.split fails if batch size is not divisible by ngpus. Error:
            #     InvalidArgumentError (see above for traceback): Number of
            #         ways to split should evenly divide the split dimension
            # xslice = tf.split(x, ngpus)[part]

            if not self._enqueue:
                return xslice

            # Did not see any benefit.
            with tf.device(dev):
                # if self._stager is None:
                stager = data_flow_ops.StagingArea(
                    dtypes=[xslice.dtype], shapes=[xslice.shape])
                stage = stager.put([xslice])
                gpucopy_ops.append(stage)
                # xslice_stage = stager.get()
            return stager.get()

        ngpus = len(gdev_list)
        if ngpus < 2:
            raise RuntimeError('Number of gpus < 2. Require two or more GPUs '
                               'for multi-gpu model parallelization.')

        model = self._smodel
        noutputs = len(self._smodel.outputs)
        global_scope = tf.get_variable_scope()
        towers = [[] for _ in range(noutputs)]
        for idev, dev in enumerate(gdev_list):
            # TODO: The last slice could cause a gradient calculation outlier
            # when averaging gradients. Maybe insure ahead of time that the
            # batch_size is evenly divisible by number of GPUs, or maybe don't
            # use the last slice.
            with tf.device(self._ps_device):
                slices = []  # multi-input case
                for ix, x in enumerate(model.inputs):
                    slice_g = Lambda(
                        slice_batch,  # lambda shape: shape,
                        # lambda shape: x.shape.as_list(),
                        name='stage_cpuSliceIn{}_Dev{}'.format(ix, idev),
                        arguments={'ngpus': ngpus, 'part': idev,
                                   'dev': dev})(x)
                    slices.append(slice_g)
                    # print('SLICE_G: {}'.format(slice_g))  # DEBUG
                # print('SLICES: {}'.format(slices))  # DEBUG

            # with tf.variable_scope('GPU_%i' % idev), \
            # tf.variable_scope(global_scope, reuse=idev > 0), \
            # tf.variable_scope('GPU_{}'.format(idev),
            #                   reuse=idev > 0) as var_scope, \
            with tf.device(dev), \
                    tf.variable_scope(global_scope, reuse=idev > 0), \
                    tf.name_scope('tower_%i' % idev):
                # NOTE: Currently not using model_creator. Did not observe
                #     any benefit in such an implementation.
                # Instantiate model under device context. More complicated.
                # Need to use optimizer synchronization in this scenario.
                # model_ = model_creator()
                # If using NCCL without re-instantiating the model then must
                # set the colocate_gradients_with_ops to False in optimizer.
                # if idev == 0:
                #     # SET STATE: Instance of serial model for checkpointing
                #     self._smodel = model_  # for ability to checkpoint

                # Handle multi-output case
                modeltower = model(slices)
                if not isinstance(modeltower, list):
                    modeltower = [modeltower]

                for imt, mt in enumerate(modeltower):
                    towers[imt].append(mt)
                    params = mt.graph._collections['trainable_variables']

                # params = model_.trainable_weights
                # params = tf.get_collection(
                #     tf.GraphKeys.TRAINABLE_VARIABLES, scope=var_scope.name)
                # params = modeltower.graph._collections['trainable_variables']
                # print('PARAMS: {}'.format(params))  # DEBUG

                self._tower_params.append(params)

        with tf.device(self._ps_device):
            # merged = Concatenate(axis=0)(towers)
            merged = [Concatenate(axis=0)(tw) for tw in towers]

        # self._enqueue_ops.append(tf.group(*gpucopy_ops))
        self._enqueue_ops += gpucopy_ops

        kwargs['inputs'] = model.inputs
        kwargs['outputs'] = merged
        super(ModelMGPU, self).__init__(*args, **kwargs)

    def compile(self, *args, **kwargs):
        '''Refer to Model.compile docstring for parameters. Override
        functionality is documented below.
        :override compile: Override Model.compile method to check for options
            that the optimizer is multi-gpu enabled, and synchronize initial
            variables.
        '''
        initsync = self._initsync
        usenccl = self._usenccl

        opt = kwargs['optimizer']
        # if isinstance(opt, str):
        if not isinstance(opt, KO.Optimizer):
            opt = KO.get(opt)
            kwargs['optimizer'] = opt

        if self._syncopt and not getattr(opt, 'ismgpu', False):
            raise RuntimeError(
                'Multi-GPU synchronization model requires a multi-GPU '
                'optimizer. Instead got: {}'.format(opt))

        opt.usenccl = usenccl

        if self._enqueue_ops:
            # Produces a warning that kwargs are ignored for Tensorflow. Patch
            # Function in tensorflow_backend to use the enqueue_ops option.
            kwargs['fetches'] = self._enqueue_ops

        super(ModelMGPU, self).compile(*args, **kwargs)

        if initsync:
            self._run_initsync()

    def _run_initsync(self):
        # tparams = [list(chain(*tp)) for tp in self._tower_params]
        tparams = self._tower_params

        # Check to prevent from unnecessarily re-initializing and
        # synchronizing, i.e. when the model loads the weights.
        for v in chain.from_iterable(tparams):
            if getattr(v, '_keras_initialized', False):
                return

        KB.manual_variable_initialization(True)
        sess = KB.get_session()
        KB.manual_variable_initialization(False)

        # glob_variables = tf.global_variables()
        # sess.run(tf.variables_initializer(glob_variables))

        # Initialize on GPU0 and sync to other GPUs
        init_op = tf.variables_initializer(tparams[0])
        # init_op = tf.variables_initializer(self._tower_params[0])
        # init_op = tf.variables_initializer(self.trainable_weights)
        sess.run(init_op)

        # Important if using model_creator. Not necessary of model instance is
        # reused in which case the model layers are shared between slices
        # and are automatically sync'd.
        sync_op = all_sync_params(tparams, self._gdev_list,
                                  usenccl=self._usenccl)
        sess.run(sync_op)

        for v in chain.from_iterable(tparams):
            v._keras_initialized = True


# Data-parallel ref: https://github.com/fchollet/keras/issues/2436
# Tower-parallel:
# ref: https://medium.com/autonomous-agents/multi-gpu-training-of-large-sparse-matrix-on-wide-neuralnetwork-cac7afc52ffe @IgnorePep8
# ref: https://gist.github.com/vvpreetham/1379cc4e208ea33ce3e615067e92fc5e
def make_parallel(serial_model, gdev_list, ps_device='/cpu:0', usenccl=False,
                  initsync=True, syncopt=False, enqueue=False,
                  model_class=ModelMGPU):
    '''Given a keras model, return an equivalent model which parallelizes
    the computation over multiple GPUs listed in the gdev_list.
    Data-Parallel:
    Each GPU gets a slice of the input batch, applies the model on that slice
    and later the outputs of the models are concatenated to a single tensor,
    hence the user sees a model that behaves the same as the original.
    If getting an out-of-memory (OOM) error when scaling the batch size by the
    number of GPUs, there might be input layer(s) in the serial model that runs
    additional special operations (such as tranformation of some sort) on the
    1st GPU as enumerated by Tensorflow. This was an observed behavior for
    Embedding layers. The workaround is to pin such layers to the CPU, or
    simply pin the instantiation of the serial mode to CPU. The parallelization
    will move the operations to GPU.
    :Example:
        if mgpu_flag:
            with tf.device('/cpu:0'):
                # define the serial model.
                model_serial = get_model_serial()
            gdev_list = get_available_gpus()
            model = make_parallel(model_serial, gdev_list)
        else:
            model = def_model_serial()
    :param Model serial_model: Serial i.e. non-multi GPU Keras model.
    :param list gdev_list: List of gpu devices i.e. ['/gpu:0', '/gpu:1', ...]
        Use function get_available_gpus to get the list of available gpus.
        This can be a list of strings or list of instances of tf.DeviceSpec.
    :param str ps_device: Parameter server device to use.
    :param bool usenccl: Use the contrib.nccl Tensorflow library for initial
        parameter synchronization and gradients averaging. Note, the model's
        usenccl option overrides the optimizers usenccl option.
        Default: False
    :param bool initsync: Synchronize initial Variables i.e. weights,
        biases, etc. Default: True
    :param bool syncopt: Synchronize gradients. Requires a multi-gpu optimizer.
        Default: False
    :param bool enqueue: Use StagingArea in the multi-GPU model. Could
        potentially speed up Host-to-Device transfers.
        Produces a warning that kwargs are ignored for Tensorflow. The
        _patch_tf_backend module mokey patches the Function in
        tensorflow_backend to use the enqueue_ops option.
        Default: False
    :param model_class: Class object to instantiate for multi-gpu models. This
        is needed when the ModelMGPU is mixed-in with other classes.
        Default: ModelMGPU
    :returns: Multi-GPU parallelized model. If ngpus < 2 then do nothing and
        return the provided serial_model.
    :rtype: ModelMGPU
    '''
    ngpus = len(gdev_list)
    if ngpus < 2:
        return serial_model  # model_creator()

    return model_class(
        serial_model=serial_model, gdev_list=gdev_list,
        ps_device=ps_device,
        enqueue=enqueue, usenccl=usenccl,
        initsync=initsync, syncopt=syncopt)



####### old version following


# # MODIFIED. Inspiration taken from the ref link below.
# # ref: https://raw.githubusercontent.com/kuza55/keras-extras/master/utils/multi_gpu.py @IgnorePep8
# # The inspirational one carried license:
# #     Apache License
# #     Version 2.0, January 2004
# # For further info refer to: https://github.com/kuza55/keras-extras
# #
# # Also used https://github.com/fchollet/keras/issues/2436 which was just
# # posted as code snippets in a forum.
# from __future__ import print_function
#
# import sys
#
# try:
#     from cStringIO import StringIO
# except ImportError:
#     # Python 3 compat.
#     from io import StringIO
#
# from itertools import chain
#
# from keras import backend as KB
# from keras.layers.core import Lambda
# from keras.models import Model
# from keras.layers.merge import Concatenate
#
# from .multi_gpu_mixin_models import ModelDataflowMixin
#
#
# if KB.backend() == 'tensorflow':
#     # Monkey patch Keras back-end to use Function with enqueue.
#     # import keras_exp._patch_tf_backend as tfbpatch
#     # tfbpatch.patch()
#     from .multi_gpu_patch_tf_backend import patch as tfbpatch
#     tfbpatch()
#
#     import tensorflow as tf
#     from tensorflow.python.client import device_lib
#
#     try:
#         from tensorflow.contrib import nccl
#         have_nccl = True
#         print('NCCL support available', file=sys.stderr)
#     except ImportError:
#         have_nccl = False
#         print('WARNING: NCCL support not available', file=sys.stderr)
#
#     from tensorflow.python.ops import data_flow_ops
#
#
# _DEBUG = False
#
# __all__ = ('get_available_gpus', 'make_parallel', 'print_mgpu_modelsummary',
#            'ModelMGPU')
#
#
# # TODO: Move to some utils module
# class Capturing(list):
#     def __enter__(self):
#         self._stdout = sys.stdout
#         sys.stdout = self._stringio = StringIO()
#         return self
#
#     def __exit__(self, *args):
#         self.extend(self._stringio.getvalue().splitlines())
#         del self._stringio    # free up some memory
#         sys.stdout = self._stdout
#
#
# def get_available_gpus(ngpus=-1):
#     '''
#     :param int ngpus: GPUs max to use. Default -1 means all gpus.
#     :returns: List of gpu devices. Ex.: ['/gpu:0', '/gpu:1', ...]
#     '''
#     local_device_protos = device_lib.list_local_devices()
#     gpus_list = [x.name for x in local_device_protos if x.device_type == 'GPU']
#     return gpus_list[:ngpus] if ngpus > -1 else gpus_list
#
#
# def print_mgpu_modelsummary(model):
#     '''Prints the summary for a multi-GPU keras model.
#     :param model: Keras model.
#     :type model: Model
#     '''
#     # print json.dumps(model.get_config(), indent=2)  # DEBUG
#     print('\nMULTI-GPU MODEL: {}'.format(model.name))
#     print(model.summary())
#     for layer in model.layers:
#         # print 'layer:', layer, '\ttype:', type(layer)
#         if isinstance(layer, Model):
#             submodel = layer
#             print('\n\tSUBMODEL SUMMARY: {}'.format(layer.name))
#             with Capturing() as msum:
#                 minfo = submodel.summary()
#             print('\t{}\n\t{}\n'.format('\n\t'.join(msum), minfo))
#
#
# def all_sync_params(tower_params, devices, usenccl=True):
#     """Assigns the params from the first tower to all others"""
#     if len(devices) == 1:
#         return tf.no_op()
#     sync_ops = []
#     if have_nccl and usenccl:
#         for param_on_devices in zip(*tower_params):
#             # print('PARAM_ON_DEVICES: {}'.format(param_on_devices))  # DEBUG
#             # Note: param_on_devices is [paramX_gpu0, paramX_gpu1, ...]
#             param0 = param_on_devices[0]
#             send_op, received_tensors = nccl.broadcast(param0, devices[1:])
#             sync_ops.append(send_op)
#             for device, param, received in zip(devices[1:],
#                                                param_on_devices[1:],
#                                                received_tensors):
#                 with tf.device(device):
#                     sync_op = param.assign(received)
#                     sync_ops.append(sync_op)
#     else:
#         params0 = tower_params[0]
#         for device, params in zip(devices, tower_params):
#             with tf.device(device):
#                 for param, param0 in zip(params, params0):
#                     sync_op = param.assign(param0.read_value())
#                     sync_ops.append(sync_op)
#
#     return tf.group(*sync_ops)
#
#
# # Data-parallel ref: https://github.com/fchollet/keras/issues/2436
# # Tower-parallel:
# # ref: https://medium.com/autonomous-agents/multi-gpu-training-of-large-sparse-matrix-on-wide-neuralnetwork-cac7afc52ffe @IgnorePep8
# # ref: https://gist.github.com/vvpreetham/1379cc4e208ea33ce3e615067e92fc5e
# def make_parallel(serial_model, gdev_list, ps_device='/cpu:0', usenccl=False,
#                   initsync=True, syncopt=False, enqueue=False):
#     '''Given a keras [model], return an equivalent model which parallelizes
#     the computation over [ngpus] GPUs.
#     Data-Parallel:
#     Each GPU gets a slice of the input batch, applies the model on that slice
#     and later the outputs of the models are concatenated to a single tensor,
#     hence the user sees a model that behaves the same as the original.
#     :param Model serial_model: Serial i.e. non-multi GPU Keras model.
#     :param list gdev_list: List of gpu devices i.e. ['/gpu:0', '/gpu:1', ...]
#         Use function get_available_gpus to get the list of available gpus.
#     :param str ps_device: Parameter server device to use.
#     :param bool usenccl: Use the contrib.nccl Tensorflow library for initial
#         parameter synchronization and gradients averaging. Note, the model's
#         usenccl option overrides the optimizers usenccl option.
#         Default: False
#     :param bool initsync: Synchronize initial Variables i.e. weights,
#         biases, etc. Default: True
#     :param bool syncopt: Synchronize gradients. Requires a multi-gpu optimizer.
#         Default: False
#     :param bool enqueue: Use StagingArea in the multi-GPU model. Could
#         potentially speed up Host-to-Device transfers.
#         Produces a warning that kwargs are ignored for Tensorflow. The
#         _patch_tf_backend module mokey patches the Function in
#         tensorflow_backend to use the enqueue_ops option.
#         Default: False
#     :returns: Multi-GPU parallelized model. If ngpus < 2 then do nothing and
#         return the provided serial_model.
#     :rtype: ModelMGPU
#     '''
#     ngpus = len(gdev_list)
#     if ngpus < 2:
#         return serial_model  # model_creator()
#
#     return ModelMGPU(serial_model=serial_model, gdev_list=gdev_list,
#                      ps_device=ps_device,
#                      enqueue=enqueue, usenccl=usenccl,
#                      initsync=initsync, syncopt=syncopt)
#
#
# # def stage(tensors):
# #     """Stages the given tensors in a StagingArea for asynchronous put/get.
# #     """
# #     stage_area = data_flow_ops.StagingArea(
# #         dtypes=[tensor.dtype for tensor in tensors],
# #         shapes=[tensor.get_shape() for tensor in tensors])
# #     put_op = stage_area.put(tensors)
# #     get_tensors = stage_area.get()
# #     if not isinstance(get_tensors, list):
# #         get_tensors = [get_tensors]
# #     # print('GET_TENSORS: {}'.format(get_tensors))  # DEBUG
# #
# #     get_tensors = [tf.reshape(gt, t.get_shape())
# #                    for (gt, t) in zip(get_tensors, tensors)]
# #     return put_op, get_tensors
#
#
# class ModelMGPU(ModelDataflowMixin, Model):
#     '''Override load and save methods of the multi-gpu model. The load and
#     save should correspond to the serial model's load and save.
#     If there are other idiosyncracies to handle for multi-gpu model case then
#     these can be handled in this subclass. A serial model should always be
#     instantiated prior to wrapping it or converting it to a multi-GPU model.
#     This multi-gpu implementation uses data-parallelism.
#     A copy-constructor is not implemented so optionally pass any additional
#     parameters besides inputes/outputs as args/kwargs to initialize the
#     multi-gpu model the same way as the serial model. Typically not needed.
#     Currently, it seems that using NCCL and synchronizing/averaging gradients
#     slows multi-gpu processing down.
#     Kwargs:
#     :param Model serial_model: Serial i.e. non-multi GPU Keras model. REQUIRED.
#     :param list gdev_list: List of gpu devices i.e. ['/gpu:0', '/gpu:1', ...]
#         Use function get_available_gpus to get the list of available gpus.
#         REQUIRED.
#     :param str ps_device: Parameter server device to use.
#     :param bool usenccl: Use the contrib.nccl Tensorflow library for initial
#         parameter synchronization and gradients averaging. Note, the models
#         usenccl option overrides the optimizers usenccl option.
#         Default: False
#         Raises RuntimeError if specified True and a non-multi-gpu optimizer is
#         passed during compile stage.
#     :param bool initsync: Synchronize initial Variables i.e. weights,
#         biases, etc. Default: True
#     :param bool syncopt: Synchronize gradients. Requires a multi-gpu optimizer.
#         Default: False
#     :param bool enqueue: Use StagingArea in the multi-GPU model. Could
#         potentially speed up Host-to-Device transfers.
#         Produces a warning that kwargs are ignored for Tensorflow. The
#         _patch_tf_backend module mokey patches the Function in
#         tensorflow_backend to use the enqueue_ops option.
#         Default: False
#     '''
#     def __init__(self, *args, **kwargs):
#         # :param model_creator: Callable that returns a serial i.e. non-multi
#         #     GPU Keras model i.e. a keras.models.Model model. REQUIRED.
#         #     Suggestion, use partial from functools to setup model_creator.
#         # try:
#         #     model_creator = kwargs.pop('model_creator')
#         # except KeyError:
#         #     raise RuntimeError('Keyword argument "model_creator" required '
#         #                        'for ModelMGPU.')
#
#         try:
#             smodel = kwargs.pop('serial_model')
#         except KeyError:
#             raise RuntimeError('Keyword argument "serial_model" required '
#                                'for ModelMGPU.')
#
#         # SET STATE: Instance of serial model for checkpointing
#         self._smodel = smodel  # model_creator()
#
#         try:
#             gdev_list = kwargs.pop('gdev_list')
#         except KeyError:
#             raise RuntimeError('Keyword argument "gdev_list" required '
#                                'for ModelMGPU.')
#         self._gdev_list = gdev_list
#
#         mname = kwargs.pop('name', self._smodel.name)
#         kwargs['name'] = mname
#
#         self._ps_device = kwargs.pop('ps_device', '/cpu:0')
#         self._initsync = kwargs.pop('initsync', True)
#         self._usenccl = kwargs.pop('usenccl', False)
#         self._syncopt = kwargs.pop('syncopt', False)
#         self._enqueue = kwargs.pop('enqueue', False)
#
#         # NOTE: To use staging have to patch keras tensorflow_backend.Function.
#         #     Function implementation in keras_exp.multigpu._patch_tf_backend
#         self._enqueue_ops = []
#
#         self._tower_params = []  # For init/sync'ing of parameters.
#         self._init_make_dataparallel(gdev_list, *args,
#                                      **kwargs)
#
#     def __getattribute__(self, attrname):
#         '''Override load and save methods to be used from the serial-model. The
#         serial-model holds references to the weights in the multi-gpu model.
#         '''
#         # return Model.__getattribute__(self, attrname)
#         if 'load' in attrname or 'save' in attrname:
#             return getattr(self._smodel, attrname)
#
#         return super(ModelMGPU, self).__getattribute__(attrname)
#
#     # ref: https://github.com/fchollet/keras/issues/2436
#     def _init_make_dataparallel(self, gdev_list, *args, **kwargs):
#         '''Uses data-parallelism to convert a serial model to multi-gpu. Refer
#         to make_parallel doc.
#         '''
#         gpucopy_ops = []
#
#         def slice_batch(x, ngpus, part, dev):
#             '''Divide the input batch into [ngpus] slices, and obtain slice
#             no. [part]. i.e. if len(x)=10, then slice_batch(x, 2, 1) will
#             return x[5:].
#             '''
#             sh = KB.shape(x)
#             L = sh[0] // ngpus
#             if part == ngpus - 1:
#                 xslice = x[part * L:]
#             else:
#                 xslice = x[part * L:(part + 1) * L]
#
#             # tf.split fails if batch size is not divisible by ngpus. Error:
#             #     InvalidArgumentError (see above for traceback): Number of
#             #         ways to split should evenly divide the split dimension
#             # xslice = tf.split(x, ngpus)[part]
#
#             if not self._enqueue:
#                 return xslice
#
#             # Did not see any benefit.
#             with tf.device(dev):
#                 # if self._stager is None:
#                 stager = data_flow_ops.StagingArea(
#                     dtypes=[xslice.dtype], shapes=[xslice.shape])
#                 stage = stager.put([xslice])
#                 gpucopy_ops.append(stage)
#                 # xslice_stage = stager.get()
#             return stager.get()
#
#         ngpus = len(gdev_list)
#         if ngpus < 2:
#             raise RuntimeError('Number of gpus < 2. Require two or more GPUs '
#                                'for multi-gpu model parallelization.')
#
#         model_ = model = self._smodel
#         global_scope = tf.get_variable_scope()
#         towers = []
#         for idev, dev in enumerate(gdev_list):
#             # TODO: The last slice could cause a gradient calculation outlier
#             # when averaging gradients. Maybe insure ahead of time that the
#             # batch_size is evenly divisible by number of GPUs, or maybe don't
#             # use the last slice.
#             with tf.device(self._ps_device):
#                 slices = []  # multi-input case
#                 for ix, x in enumerate(model.inputs):
#                     slice_g = Lambda(
#                         slice_batch,  # lambda shape: shape,
#                         lambda shape: x.shape.as_list(),
#                         name='stage_cpuSliceIn{}_Dev{}'.format(ix, idev),
#                         arguments={'ngpus': ngpus, 'part': idev,
#                                    'dev': dev})(x)
#                     slices.append(slice_g)
#                     # print('SLICE_G: {}'.format(slice_g))  # DEBUG
#                 # print('SLICES: {}'.format(slices))  # DEBUG
#
#             # with tf.variable_scope('GPU_%i' % idev), \
#             # tf.variable_scope(global_scope, reuse=idev > 0), \
#             # tf.variable_scope('GPU_{}'.format(idev),
#             #                   reuse=idev > 0) as var_scope, \
#             with tf.device(dev), \
#                     tf.variable_scope(global_scope, reuse=idev > 0), \
#                     tf.name_scope('tower_%i' % idev):
#                 # NOTE: Currently not using model_creator. Did not observe
#                 #     any benefit in such an implementation.
#                 # Instantiate model under device context. More complicated.
#                 # Need to use optimizer synchronization in this scenario.
#                 # model_ = model_creator()
#                 # If using NCCL without re-instantiating the model then must
#                 # set the colocate_gradients_with_ops to False in optimizer.
#                 # if idev == 0:
#                 #     # SET STATE: Instance of serial model for checkpointing
#                 #     self._smodel = model_  # for ability to checkpoint
#
#                 modeltower = model_(slices)
#                 towers.append(modeltower)
#
#                 # params = model_.trainable_weights
#                 # params = tf.get_collection(
#                 #     tf.GraphKeys.TRAINABLE_VARIABLES, scope=var_scope.name)
#                 params = modeltower.graph._collections['trainable_variables']
#                 # print('PARAMS: {}'.format(params))  # DEBUG
#
#                 self._tower_params.append(params)
#
#         with tf.device(self._ps_device):
#             merged = Concatenate(axis=0)(towers)
#             # print('MERGED: {}'.format(merged))  # DEBUG
#
#         # self._enqueue_ops.append(tf.group(*gpucopy_ops))
#         self._enqueue_ops += gpucopy_ops
#
#         kwargs['inputs'] = model.inputs
#         kwargs['outputs'] = merged
#         super(ModelMGPU, self).__init__(*args, **kwargs)
#
#     def compile(self, *args, **kwargs):
#         '''Refer to Model.compile docstring for parameters. Override
#         functionality is documented below.
#         :override compile: Override Model.compile method to check for options
#             that the optimizer is multi-gpu enabled, and synchronize initial
#             variables.
#         '''
#         initsync = self._initsync
#         usenccl = self._usenccl
#
#         opt = kwargs['optimizer']
#         if self._syncopt and not getattr(opt, 'ismgpu', False):
#             raise RuntimeError(
#                 'Multi-GPU synchronization model requires a multi-GPU '
#                 'optimizer. Instead got: {}'.format(opt))
#
#         opt.usenccl = usenccl
#
#         if self._enqueue_ops:
#             # Produces a warning that kwargs are ignored for Tensorflow. Patch
#             # Function in tensorflow_backend to use the enqueue_ops option.
#             kwargs['enqueue_ops'] = self._enqueue_ops
#
#         super(ModelMGPU, self).compile(*args, **kwargs)
#
#         if initsync:
#             self._run_initsync()
#
#     def _run_initsync(self):
#         # tparams = [list(chain(*tp)) for tp in self._tower_params]
#         tparams = self._tower_params
#
#         # Check to prevent from unnecessarily re-initializing and
#         # synchronizing, i.e. when the model loads the weights.
#         for v in chain.from_iterable(tparams):
#             if getattr(v, '_keras_initialized', False):
#                 return
#
#         KB.manual_variable_initialization(True)
#         sess = KB.get_session()
#         KB.manual_variable_initialization(False)
#
#         # glob_variables = tf.global_variables()
#         # sess.run(tf.variables_initializer(glob_variables))
#
#         # Initialize on GPU0 and sync to other GPUs
#         init_op = tf.variables_initializer(tparams[0])
#         # init_op = tf.variables_initializer(self._tower_params[0])
#         # init_op = tf.variables_initializer(self.trainable_weights)
#         sess.run(init_op)
#
#         # Important if using model_creator. Not necessary of model instance is
#         # reused in which case the model layers are shared between slices
#         # and are automatically sync'd.
#         sync_op = all_sync_params(tparams, self._gdev_list,
#                                   usenccl=self._usenccl)
#         sess.run(sync_op)
#
#         for v in chain.from_iterable(tparams):
#             v._keras_initialized = True




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for training and validating NN's, as well as evaluating them.
"""

from inspect import signature
import h5py
from contextlib import ExitStack
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from orcanet.utilities.visualization import (
    plot_activations, plot_weights)
from orcanet.logging import BatchLogger

# for debugging
# from tensorflow.python import debug as tf_debug
# K.set_session(tf_debug.LocalCLIDebugWrapperSession(tf.Session()))


def train_model(orga, model, epoch, batch_logger=False):
    """
    Train a model on one file and return the history.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras model
        A compiled keras model.
    epoch : tuple
        Current epoch and the no of the file to train on.
    batch_logger : bool
        Use the orcanet batchlogger to log the training.

    Returns
    -------
    history : dict
        The history of the training on this file. A record of training
        loss values and metrics values.

    """
    files_dict = orga.io.get_file("train", epoch[1])

    if orga.cfg.n_events is not None:
        # TODO Can throw an error if n_events is larger than the file
        f_size = orga.cfg.n_events  # for testing purposes
    else:
        f_size = orga.io.get_file_sizes("train")[epoch[1] - 1]

    callbacks = []
    if batch_logger:
        callbacks.append(BatchLogger(orga, epoch))
    if orga.cfg.callback_train is not None:
        try:
            callbacks.extend(orga.cfg.callback_train)
        except TypeError:
            callbacks.append(orga.cfg.callback_train)

    training_generator = hdf5_batch_generator(
        orga, files_dict, f_size=f_size,
        zero_center=orga.cfg.zero_center_folder is not None,
        shuffle=orga.cfg.shuffle_train)

    history = model.fit_generator(
        training_generator,
        steps_per_epoch=int(f_size / orga.cfg.batchsize),
        verbose=orga.cfg.verbose_train,
        max_queue_size=orga.cfg.max_queue_size,
        callbacks=callbacks,
        initial_epoch=epoch[0],
        epochs=epoch[0] + 1,
    )

    # get a dict with losses and metrics
    # only trained for one epoch, so value is list of len 1
    history = {key: value[0] for key, value in history.history.items()}
    return history


def validate_model(orga, model):
    """
    Validates a model on all validation files and return the history.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras model
        A compiled keras model.

    Returns
    -------
    history : dict
        The history of the validation on all files. A record of validation
        loss values and metrics values.

    """
    # One history for each val file
    histories = []
    f_sizes = orga.io.get_file_sizes("val")

    for i, files_dict in enumerate(orga.io.yield_files("val")):
        f_size = f_sizes[i]
        if orga.cfg.n_events is not None:
            f_size = orga.cfg.n_events  # for testing purposes

        val_generator = hdf5_batch_generator(
            orga, files_dict, f_size=f_size,
            zero_center=orga.cfg.zero_center_folder is not None)

        history_file = model.evaluate_generator(
            val_generator,
            steps=int(f_size / orga.cfg.batchsize),
            max_queue_size=orga.cfg.max_queue_size,
            verbose=orga.cfg.verbose_val)

        histories.append(history_file)

    # average over all val files if necessary
    history = [sum(col) / float(len(col)) for col in zip(*histories)] \
        if len(histories) > 1 else histories[0]

    # This history is just a list, not a dict like with fit_generator
    # so transform to dict
    history = dict(zip(model.metrics_names, history))

    return history


def hdf5_batch_generator(orga, files_dict, f_size=None, zero_center=False,
                         yield_mc_info=False, shuffle=False):
    """
    Yields batches of input data from h5 files.

    This will go through one file, or multiple files in parallel, and yield
    one batch of data, which can then be used as an input to a model.
    Since multiple filepaths can be given to read out in parallel,
    this can also be used for models with multiple inputs.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    files_dict : dict
        Pathes of the files to train on.
        Keys: The name of every input (from the toml list file, can be multiple).
        Values: The filepath of a single h5py file to read samples from.
    f_size : int or None
        Specifies the number of samples to be read from the .h5 file.
        If none, the whole .h5 file will be used.
    zero_center : bool
        Whether to use zero centering.
        Requires orga.zero_center_folder to be set.
    yield_mc_info : bool
        Specifies if mc-infos (y_values) should be yielded as well. The
        mc-infos are used for evaluation after training and testing is finished.
    shuffle : bool
        Randomize the order in which batches are read from the file.
        Significantly reduces read out speed.

    Yields
    ------
    xs : dict
        Data for the model train on.
            Keys : str  The name(s) of the input layer(s) of the model.
            Values : ndarray    A batch of samples for the corresponding input.
    ys : dict
        Labels for the model to train on.
            Keys : str  The name(s) of the output layer(s) of the model.
            Values : ndarray    A batch of labels for the corresponding output.
    mc_info : ndarray, optional
        Mc info from the file. Only yielded if yield_mc_info is True.

    """
    batchsize = orga.cfg.batchsize
    # name of the datagroups in the file
    samples_key = orga.cfg.key_samples
    mc_key = orga.cfg.key_labels

    # If the batchsize is larger than the f_size, make batchsize smaller
    # or nothing would be yielded
    if f_size is not None:
        if f_size < batchsize:
            batchsize = f_size

    if orga.cfg.label_modifier is not None:
        label_modifier = orga.cfg.label_modifier
    else:
        assert orga._auto_label_modifier is not None, \
            "Auto label modifier has not been set up (can be done with " \
            "nn_utilities.get_auto_label_modifier)"
        label_modifier = orga._auto_label_modifier

    # get xs_mean or load/create if not stored yet
    if zero_center:
        xs_mean = orga.get_xs_mean()
    else:
        xs_mean = None

    with ExitStack() as stack:
        # a dict with the names of list inputs as keys, and the opened
        # h5 files as values
        files = {}
        file_lengths = []
        # open the files and make sure they have the same length
        for input_key in files_dict:
            files[input_key] = stack.enter_context(
                h5py.File(files_dict[input_key], 'r'))
            file_lengths.append(len(files[input_key][samples_key]))

        if not file_lengths.count(file_lengths[0]) == len(file_lengths):
            raise ValueError("All data files must have the same length! "
                             "Yours have:\n " + str(file_lengths))

        if f_size is None:
            f_size = file_lengths[0]
        total_no_of_batches = int(np.ceil(f_size/batchsize))
        # positions of the samples in the file
        sample_pos = np.arange(total_no_of_batches) * batchsize
        if shuffle:
            np.random.shuffle(sample_pos)
        # append some samples due to preloading by the fit_generator method
        sample_pos = np.append(sample_pos, sample_pos[:orga.cfg.max_queue_size])

        for sample_n in sample_pos:
            # A dict with every input name as key, and a batch of data as values
            xs = {}
            # Read one batch of samples from the files and zero center
            for input_key in files:
                xs[input_key] = files[input_key][samples_key][
                                sample_n: sample_n + batchsize]
                if xs_mean is not None:
                    xs[input_key] = np.subtract(xs[input_key],
                                                xs_mean[input_key])
            # Get labels for the nn. Since the labels are hopefully the same
            # for all the files, use the ones from the first TODO add check
            y_values = list(files.values())[0][mc_key][
                       sample_n:sample_n + batchsize]

            # Modify the samples and labels before feeding them into the network
            if orga.cfg.sample_modifier is not None:
                xs = orga.cfg.sample_modifier(xs)

            ys = label_modifier(y_values)

            if not yield_mc_info:
                yield xs, ys
            else:
                yield xs, ys, y_values


def get_learning_rate(epoch, user_lr, no_train_files):
    """
    Get the learning rate for the current epoch and file number.

    Parameters
    ----------
    epoch : tuple
        Epoch and file number.
    user_lr : float or tuple or function.
        The user input for the lr.
    no_train_files : int
        How many train files there are in total.

    Returns
    -------
    lr : float
        The learning rate.

    Raises
    ------
    AssertionError
        If the type of the user_lr is not right.

    """
    error_msg = "The learning rate must be either a float, a tuple of " \
                "two floats or a function."
    try:
        # Float => Constant LR
        lr = float(user_lr)
        return lr
    except (ValueError, TypeError):
        pass

    try:
        # List => Exponentially decaying LR
        length = len(user_lr)
        lr_init = float(user_lr[0])
        lr_decay = float(user_lr[1])
        if length != 2:
            raise TypeError("{} (Your tuple has length {})".format(error_msg,
                                                                   len(user_lr)))

        lr = lr_init * (1 - lr_decay) ** ((epoch[1]-1) + (epoch[0]-1) * no_train_files)
        return lr
    except (ValueError, TypeError):
        pass

    try:
        # Callable => User defined function
        n_params = len(signature(user_lr).parameters)
        if n_params != 2:
            raise TypeError("A custom learning rate function must have two "
                            "input parameters: The epoch and the file number. "
                            "(yours has {})".format(n_params))
        lr = user_lr(epoch[0], epoch[1])
        return lr
    except (ValueError, TypeError):
        raise TypeError("{} (You gave {} of type {}) ".format(
            error_msg, user_lr, type(user_lr)))


def save_actv_wghts_plot(orga, model, epoch, samples=1):
    """
    Plots the weights of a model and the activations for samples from
    the validation set to one .pdf file each.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : ks.models.Model
        The model to do the predictions with.
    epoch : tuple
        Current epoch and fileno.
    samples : int
        Number of samples to make the plot for.

    """
    plt.ioff()

    file = next(orga.io.yield_files("val"))
    generator = hdf5_batch_generator(
        orga, file, f_size=samples,
        zero_center=orga.cfg.zero_center_folder is not None,
        yield_mc_info=True)
    xs, ys, y_values = next(generator)

    pdf_name_act = "{}/activations_epoch_{}_file_{}.pdf".format(
        orga.io.get_subfolder("activations", create=True), epoch[0], epoch[1])

    with PdfPages(pdf_name_act) as pdf:
        for layer in model.layers:
            fig = plot_activations(model, xs, layer.name, mode='test')
            pdf.savefig(fig)
            plt.close(fig)

    pdf_name_wght = "{}/weights_epoch_{}_file_{}.pdf".format(
        orga.io.get_subfolder("activations", create=True), epoch[0], epoch[1])

    with PdfPages(pdf_name_wght) as pdf:
        for layer in model.layers:
            fig = plot_weights(model, layer.name)
            if fig is not None:
                pdf.savefig(fig)
                plt.close(fig)


def make_model_prediction(orga, model, eval_filename, samples=None):
    """
    Let a model predict on all validation samples, and save it as a h5 file.

    Per default, the h5 file will contain a datagroup mc_info straight from
    the given files, as well as two datagroups per output layer of the network,
    which have the labels and the predicted values in them as numpy arrays,
    respectively.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : ks.model.Model
        Trained Keras model of a neural network.
    eval_filename : str
        Name and path of the h5 file.
    samples : int or None
        Number of events that should be predicted.
        If samples=None, the whole file will be used.

    """
    batchsize = orga.cfg.batchsize
    compression = ("gzip", 1)
    file_sizes = orga.io.get_file_sizes("val")
    total_file_size = sum(file_sizes)
    datagroups_created = False

    with h5py.File(eval_filename, 'w') as h5_file:
        # For every val file set (one set can have multiple files if
        # the model has multiple inputs):
        for f_number, files_dict in enumerate(orga.io.yield_files("val")):
            file_size = file_sizes[f_number]
            generator = hdf5_batch_generator(
                orga, files_dict,
                zero_center=orga.cfg.zero_center_folder is not None,
                yield_mc_info=True)

            if samples is None:
                steps = int(file_size / batchsize)
                if file_size % batchsize != 0:
                    # add a smaller step in the end
                    steps += 1
            else:
                steps = int(samples / batchsize)

            for s in range(steps):
                if s % 100 == 0:
                    print('Predicting in step {} on file {}'.format(s, f_number))
                # y_true is a dict of ndarrays, mc_info is a structured
                # array, y_pred is a list of ndarrays
                xs, y_true, mc_info = next(generator)

                y_pred = model.predict_on_batch(xs)
                if not isinstance(y_pred, list):
                    # if only one output, transform to a list
                    y_pred = [y_pred]
                # transform y_pred to dict
                y_pred = {out: y_pred[i] for i, out in enumerate(model.output_names)}

                if orga.cfg.dataset_modifier is None:
                    datasets = get_datasets(mc_info, y_true, y_pred)
                else:
                    datasets = orga.cfg.dataset_modifier(mc_info, y_true, y_pred)

                # TODO maybe add attr to data, like used files or orcanet version number?
                if not datagroups_created:
                    for dataset_name, data in datasets.items():
                        maxshape = (total_file_size,) + data.shape[1:]
                        chunks = True  # (batchsize,) + data.shape[1:]
                        h5_file.create_dataset(
                            dataset_name, data=data, maxshape=maxshape,
                            chunks=chunks, compression=compression[0],
                            compression_opts=compression[1])
                        datagroups_created = True
                else:
                    for dataset_name, data in datasets.items():
                        # append data at the end of the dataset
                        h5_file[dataset_name].resize(
                            h5_file[dataset_name].shape[0] + data.shape[0], axis=0)
                        h5_file[dataset_name][-data.shape[0]:] = data


def get_datasets(mc_info, y_true, y_pred):
    """
    Get the dataset names and numpy array contents.

    Every output layer will get one dataset each for both the label and
    the prediction. E.g. if your model has an output layer called "energy",
    the datasets "label_energy" and "pred_energy" will be made.

    Parameters
    ----------
    mc_info : ndarray
        A structured array containing infos for every event, right from
        the input files.
    y_true : dict
        The labels for each output layer of the network.
    y_pred : dict
        The predictions of each output layer of the network.

    Returns
    -------
    datasets : dict
        Keys are the name of the datagroups, values the content in the
        form of numpy arrays.

    """
    datasets = dict()
    datasets["mc_info"] = mc_info
    for out_layer_name in y_true:
        datasets["label_" + out_layer_name] = y_true[out_layer_name]
    for out_layer_name in y_pred:
        datasets["pred_" + out_layer_name] = y_pred[out_layer_name]
    return datasets



# -*- coding: utf-8 -*-
"""Wide Residual Network models for Keras.
Loosely based on https://github.com/titu1994/Wide-Residual-Networks/blob/master/wide_residual_network.py

# Reference
- [Wide Residual Networks](https://arxiv.org/abs/1605.07146)
"""
from keras.models import Model
from keras.layers import Input, Add, Activation, Dropout, Flatten, Dense
from keras.layers.convolutional import Convolution2D, Convolution3D, \
    AveragePooling2D, AveragePooling3D
from keras.layers.normalization import BatchNormalization
from keras import backend as K

from scraps.old_model_scripts.file_dump import get_dimensions_encoding


def decode_input_dimensions(n_bins, batchsize, swap_4d_channels):
    """
    Returns the input dimensions (e.g. batchsize x 11 x 13 x 18 x Channels for 3D)
    and appropriate strides and avg_pooling sizes depending on the projection type.
    :param tuple n_bins: Number of bins (x,y,z,t) of the data.
    :param int batchsize: Batchsize of the fed data.
    :param None/str swap_4d_channels: For 3.5D nets, specifies if the default channel (t) should be swapped with another dim.
    :return: tuple input_dim: dimensions tuple for 2D, 3D or 4D data (ints).
    :return: [tuple, tuple] strides: Strides of the WRN.
    :return: tuple average_pooling_size: Avg_pooling size of the WRN.
    """
    input_dim = get_dimensions_encoding(n_bins, batchsize) # includes batchsize

    if n_bins.count(1) == 2: # 2d case
        dim = 2

        if n_bins[0] == 1 and n_bins[3] == 1:
            print('Using a Wide ResNet with YZ projection')
            strides = [(1,1), (1,1), (2,2)]
            average_pooling_size = (7,9)

        elif n_bins[0] == 1 and n_bins[1] == 1:
            print('Using a Wide ResNet with ZT projection')
            strides = [(1,1), (1,2), (2,2)]
            average_pooling_size = (9,13)

        else:
            raise IndexError('No suitable 2D projection found in decode_input_dimensions().'
                             'Please add the projection type with the strides and the average_pooling_size to the function.')

    elif n_bins.count(1) == 1: # 3d case
        dim = 3

        if n_bins[1] == 1:
            print('Using a Wide ResNet with XZT projection')
            strides = [(1,1,1), (1,1,2), (2,2,2)]
            average_pooling_size = (6,9,13)

        elif n_bins[3] == 1:
            print('Using a Wide ResNet with XYZ projection')
            strides = [(1,1,1), (1,1,1), (2,2,2)]
            average_pooling_size = (6,7,9)

        elif n_bins[0] == 1:
            print('Using a Wide ResNet with YZT projection')
            strides = [(1,1,1), (1,1,2), (2,2,2)]
            average_pooling_size = (7,9,13)

        else:
            raise IndexError('No suitable 3D projection found in decode_input_dimensions().'
                             'Please add the projection type with the strides and the average_pooling_size to the function.')

    elif n_bins.count(1) == 0:
        dim = 3

        if swap_4d_channels is None:
            print('Using a WRN 3.5D CNN with XYZ data and T channel information.')
            #strides = [(1,1,1), (1,1,1), (2,2,2)]
            strides = [(1, 1, 1), (2, 2, 2)]
            average_pooling_size = (6,7,9)

        elif swap_4d_channels == 'yzt-x':
            print('Using a WRN 3.5D CNN with YZT data and X channel information.')
            #strides = [(1,1,1), (1,1,2), (2,2,2)]
            strides = [(1,1,2), (2,2,2)]
            average_pooling_size = (7,9,13)
            input_dim = (input_dim[0], input_dim[2], input_dim[3], input_dim[4], input_dim[1]) # [bs,y,z,t,x]

        else:
            raise IOError('3.5D projection types other than XYZ-T and YZT-X are not yet supported.'
                          'Please add the max_pool_sizes dict in the function by yourself.')

    else:
        raise IOError('Data types other than 2D, 3D or 4D (3.5D actually) are not yet supported. '
                      'Please specify a 2D, 3D or 4D n_bins tuple.')

    return dim, input_dim, strides, average_pooling_size


def create_wide_residual_network(n_bins, n=1, k=8, dropout=0.0, k_size=3, verbose=True, swap_4d_channels=None):
    """
    Creates a 2D or 3D Wide Residual Network with specified parameters.
    The torch implementation from the paper differs slightly (change default arguments in BatchNorm):
    - BatchNormalization(axis=channel_axis, momentum=0.1, epsilon=1e-5, gamma_initializer='uniform')(x)
    :param tuple n_bins: Number of bins (x,y,z,t) of the data that will be fed to the network.
    :param int nb_classes: Number of output classes.
    :param int n: Depth of the network. Compute n = (N - 4) / 6.
                  Example : For a depth of 16, N = 16, n = (16 - 4) / 6 = 2
                  Example2: For a depth of 28, N = 28, n = (28 - 4) / 6 = 4
                  Example3: For a depth of 40, N = 40, n = (40 - 4) / 6 = 6
    :param int k: Width of the network (gets multiplied by the number of filters for each convolution).
    :param float dropout: Adds dropout if value is greater than 0.0.
    :param int k_size: Kernel size that should be used, same for each dimension.
    :param bool verbose: Debug info to describe the created WRN.
    :param None/str swap_4d_channels: For 3.5D nets, specifies if the default channel (t) should be swapped with another dim.
    :return: Model model: Keras WRN model.
    """
    nb_classes = 2
    channel_axis = 1 if K.image_data_format() == "channels_first" else -1

    # TODO Batchsize has to be given to decode_input_dimensions_vgg, but is not used for constructing the model.
    # For now: Just use some random value.
    batchsize = 64

    dim, input_dim, strides, avg_pool_size = decode_input_dimensions(n_bins, batchsize, swap_4d_channels)  # includes batchsize
    average_pooling_nd = AveragePooling2D if dim==2 else AveragePooling3D

    input_layer = Input(shape=input_dim[1:], dtype=K.floatx()) # batch_shape=input_dim

    #x = BatchNormalization(axis=channel_axis)(input_layer) # can be used after input in order to omit mean substraction. Care: Change initial conv param to x
    x = initial_conv(input_layer, 64, dim, dropout=dropout, k_size=k_size, channel_axis=channel_axis)
    x = expand_conv(x, 64, dim, k=k, dropout=dropout, k_size=k_size, strides=strides[0], channel_axis=channel_axis)
    nb_conv = 10 # 1x initial_conv + 3x expand_conv = 1x1 + 3*3 = 10

    for i in range(n):
        x = conv_block(x, 64, dim, k=k, dropout=dropout, k_size=k_size, channel_axis=channel_axis)
        nb_conv += 2

    x = BatchNormalization(axis=channel_axis)(x)
    x = Activation('relu')(x)

    if dropout > 0.0: x = Dropout(dropout)(x)

    x = expand_conv(x, 128, dim, k=k, dropout=dropout, strides=strides[1], channel_axis=channel_axis)

    for i in range(n):
        x = conv_block(x, 128, dim, k=k, dropout=dropout, k_size=k_size, channel_axis=channel_axis)
        nb_conv += 2

    x = BatchNormalization(axis=channel_axis)(x)
    x = Activation('relu')(x)

    # x = expand_conv(x, 64, dim, k=k, dropout=dropout, strides=strides[2], channel_axis=channel_axis)
    #
    # for i in range(n):
    #     x = conv_block(x, 64, dim, k=k, dropout=dropout, k_size=k_size, channel_axis=channel_axis)
    #     nb_conv += 2
    #
    # x = BatchNormalization(axis=channel_axis)(x)
    # x = Activation('relu')(x)

    if dropout > 0.0: x = Dropout(dropout)(x)

    x = average_pooling_nd(avg_pool_size)(x) # use global average pooling instead of fully connected
    x = Flatten()(x)

    # could also be transformed to one neuron -> binary_crossentropy + sigmoid instead of 2 neurons -> cat._crossentropy + softmax
    x = Dense(nb_classes, activation='softmax', kernel_initializer='he_normal')(x) # actually linear in the paper, could also be transformed to one neuron

    model = Model(input_layer, x)

    if verbose: print("Wide Residual Network-%d-%d created." % (nb_conv, k))
    return model


def initial_conv(input_layer,n_filters, dim, dropout=0.0, k_size=3, channel_axis=-1):
    """
    Initial convolution prior to the ResNet blocks (2D/3D).
    C-B-A
    :param ks.layers.Input input_layer: Keras Input layer (tensor) that specifies the shape of the input data.
    :param int n_filters: Number of filters used for the initial convolution.
    :param int dim: 2D or 3D block.
    :param float dropout: Adds dropout if >0.
    :param int k_size: Kernel size that should be used.
    :param int channel_axis: the channel axis that the BatchNorm layer should be applied on.
    :return: x: Resulting output tensor (model).
    """
    if dim not in (2,3): raise ValueError('dim must be equal to 2 or 3.')
    convolution_nd = Convolution2D if dim==2 else Convolution3D

    x = convolution_nd(n_filters, (k_size,) * dim, padding='same', kernel_initializer='he_normal', use_bias=False)(input_layer) # TODO probably more filters, standard=16

    x = BatchNormalization(axis=channel_axis)(x)
    x = Activation('relu')(x)

    if dropout > 0.0: x = Dropout(dropout)(x)

    return x


def expand_conv(ip, n_filters, dim, k=1, dropout=0.0, k_size=3, strides=None, channel_axis=-1):
    """
    Intermediate convolution block that expands the number of features (increase number of filters, i.e. 16->32).
    2D/3D.
    (C-B-A-C) + (C) -> M
    :param ip: Keras functional layer instance that is used as the starting point of this convolutional block.
    :param int n_filters: Number of filters used for each convolution.
    :param int dim: 2D or 3D block.
    :param int k: Width of the convolution, multiplicative factor to "n_filters".
    :param float dropout: Adds dropout if >0.
    :param int k_size: Kernel size which is used for all three dimensions.
    :param tuple(int) strides: Strides of the convolutional layers.
    :param int channel_axis: the channel axis that the BatchNorm layer should be applied on.
    :return: m: Keras functional layer instance where the last layer is the merge.
    """
    if dim not in (2, 3): raise ValueError('dim must be equal to 2 or 3.')
    if strides is None: strides = (1,) * dim
    convolution_nd = Convolution2D if dim == 2 else Convolution3D

    x = convolution_nd(n_filters * k, (k_size,) * dim, padding='same', strides=strides, kernel_initializer='he_normal', use_bias=False)(ip)

    x = BatchNormalization(axis=channel_axis)(x)
    x = Activation('relu')(x)

    if dropout > 0.0: x = Dropout(dropout)(x)

    x = convolution_nd(n_filters * k, (k_size,) * dim, padding='same', kernel_initializer='he_normal', use_bias=False)(x)
    skip = convolution_nd(n_filters * k, (1,) * dim, padding='same', strides=strides, kernel_initializer='he_normal', use_bias=False)(ip)

    if dropout > 0.0: skip = Dropout(dropout)(skip) #

    m = Add()([x, skip])

    return m


def conv_block(ip, n_filters, dim, k=1, dropout=0.0, k_size=3, channel_axis=-1):
    """
    ResNet block (2D/3D).
    (B-A-C-B-A-C) + (ip) = M
    :param ip: Keras functional layer instance that is used as the starting point of this convolutional block.
    :param int n_filters: Number of filters used for each convolution.
    :param int dim: 2D or 3D block.
    :param int k: Width of the convolution, multiplicative factor to "n_filters".
    :param float dropout: Adds dropout if >0.
    :param int k_size: Kernel size which is used for all three dimensions.
    :param int channel_axis: the channel axis that the BatchNorm layer should be applied on.
    :return: m: Keras functional layer instance where the last layer is the merge.
    """
    if dim not in (2, 3): raise ValueError('dim must be equal to 2 or 3.')
    convolution_nd = Convolution2D if dim == 2 else Convolution3D

    init = ip # TODO useless?

    x = BatchNormalization(axis=channel_axis)(ip)
    x = Activation('relu')(x)

    if dropout > 0.0: x = Dropout(dropout)(x) ##

    x = convolution_nd(n_filters * k, (k_size,) * dim, padding='same', kernel_initializer='he_normal', use_bias=False)(x)

    x = BatchNormalization(axis=channel_axis)(x)
    x = Activation('relu')(x)

    if dropout > 0.0: x = Dropout(dropout)(x)

    x = convolution_nd(n_filters * k, (k_size,) * dim, padding='same', kernel_initializer='he_normal', use_bias=False)(x)

    if dropout > 0.0: init = Dropout(dropout)(init) #

    m = Add()([init, x])
    return m


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for making plots for background classifiers (muon/random_noise/neutrinos).
"""

from matplotlib import pyplot as plt
import numpy as np
from orcanet_contrib.plotting.utils import get_event_selection_mask


def make_prob_hists_bg_classifier(pred_file, savefolder, cuts=None):
    """
    Function that makes plots for the reco probability distributions of the background classifier.

    Makes one plot for each probability output (prob_muon, prob_random_noise, prob_neutrino)

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the background classification predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    """
    def configure_and_save_plot(bg_class, savefolder):
        """
        Configure and save a mpl plot with GridLines, Logscale etc. for the background classifier probability plots.
        Parameters
        ----------
        bg_class : str
            A string which specifies the probability class: (muon, random_noise, neutrino).
        savefolder : str
            Path to the directory, where the plot should be saved.

        """
        axes.legend(loc='upper center', ncol=1)
        plt.grid(True, zorder=0, linestyle='dotted')
        plt.yscale('log')

        x_ticks_major = np.arange(0, 1.1, 0.1)
        plt.xticks(x_ticks_major)
        plt.minorticks_on()

        plt.xlabel('OrcaNet ' + bg_class + ' probability')
        plt.ylabel('Normed Quantity')
        title = plt.title('Probability to be classified as ' + bg_class)
        title.set_position([.5, 1.04])

        plt.savefig(savefolder + '/prob_' + bg_class + '.pdf')
        plt.savefig(savefolder + '/prob_' + bg_class + '.png', dpi=600)

    fig, axes = plt.subplots()

    #bg_classes = ['muon', 'random_noise', 'neutrino']
    bg_classes = ['not_neutrino', 'neutrino']
    for bg_class in bg_classes:
        prob_class = pred_file['pred']['prob_' + bg_class]
        ptype, is_cc = pred_file['mc_info']['particle_type'], pred_file['mc_info']['is_cc']

        if cuts is not None:
            assert isinstance(cuts, str)
            evt_sel_mask = get_event_selection_mask(pred_file['mc_info'], cut_name=cuts)
            prob_class = prob_class[evt_sel_mask]
            ptype, is_cc = ptype[evt_sel_mask], is_cc[evt_sel_mask]

        make_prob_hists_for_class(prob_class, ptype, is_cc, axes)
        configure_and_save_plot(bg_class, savefolder)
        plt.cla()


def select_class(class_name, ptype, is_cc):
    """
    Returns a boolean array which specifies, which rows in the ptype & is_cc 1d arrays belong to the class.

    Parameters
    ----------
    class_name : str
        String that specifies the class that should be selected for the boolean flag output.
    ptype : ndarray(ndim=1)
        Array with particle_types of some events.
    is_cc : ndarray(ndim=1)
        Array with is_cc of some events.

    Returns
    -------
    is_class : ndarray(ndim=1)
        Boolean flag, which specifies if each row belongs to the class specified by the class_name.

    """
    if class_name == 'mupage':
        is_class = np.abs(ptype) == 13
    elif class_name == 'random_noise':
        is_class = np.abs(ptype == 0)
    elif class_name == 'neutrino':
        is_class = np.logical_or.reduce((np.abs(ptype == 12), np.abs(ptype == 14), np.abs(ptype == 16)))
    elif class_name == 'neutrino_track':
        is_class = np.logical_and(np.abs(ptype == 14), is_cc == 1)
    elif class_name == 'neutrino_shower':
        is_muon_nc = np.logical_and(np.abs(ptype == 14), is_cc == 0)
        is_class = np.logical_or.reduce((np.abs(ptype == 12), np.abs(ptype == 16), is_muon_nc))
    else:
        raise ValueError('The class ' + str(class_name) + ' is not available.')

    return is_class


def make_prob_hists_for_class(prob_class, ptype, is_cc, axes):
    """
    Makes a mpl hist plot of the probabilities for a certain probability class (either muon, random_noise or neutrino).

    Plots the probabilities for this prob_class for all events (muon, random_noise, neutrinos)

    Parameters
    ----------
    prob_class : ndarray(ndim=1)
        Reco probabilities for events for a certain probability class (muon, random_noise, neutrinos).
    ptype : ndarray(ndim=1)
        Array with particle_types of the same events as in the prob_class.
    is_cc : ndarray(ndim=1)
        Array with is_cc of the same events as in the prob_class.
    axes : mpl.axes
        Axes object that refers to ax of an existing plt.sublots object.

    """
    is_mupage = select_class('mupage', ptype, is_cc)
    is_random_noise = select_class('random_noise', ptype, is_cc)
    is_neutrino = select_class('neutrino', ptype, is_cc)

    axes.hist(prob_class[is_mupage], density=True, bins=40, range=(0, 1), color='b', label='mupage', histtype='step', linestyle='-', zorder=3)
    axes.hist(prob_class[is_random_noise], density=True, bins=40, range=(0, 1), color='r', label='random_noise', histtype='step', linestyle='-', zorder=3)
    axes.hist(prob_class[is_neutrino], density=True, bins=40, range=(0, 1), color='saddlebrown', label='neutrino', histtype='step', linestyle='-', zorder=3)



"""
For testing how long generators take.

Will show the average time per batch a generator takes, as well as the average training time per batch of a model.

Shuffle test:
-------------
Generator shuffles the order in which data gets read out.
Tested on 5000 batches each of (64, 11, 13, 18, 60) xyzt data with the vgg model with 3,000,000 params.
Results:

            Read-time per batch (s)    Train time per batch (s)
No Shuffle        0.07755                    0.2017
Shuffle           0.1276                     0.198

+65% due to shuffle! Still lower than train time though.

"""
import os
import time
import numpy as np
import h5py
from orcanet.core import Organizer
from model_builder import build_nn_model


def generate_batches_from_hdf5_file_tweak(cfg, files_dict, f_size=None, zero_center_image=None, yield_mc_info=False, shuffle=False, batches_at_once=10):
    """
    Yields batches of input data from h5 files.

    This will go through one file, or multiple files in parallel, and yield one batch of data, which can then
    be used as an input to a model. Since multiple filepaths can be given to read out in parallel,
    this can also be used for models with multiple inputs.
    # TODO Is reading n batches at once and yielding them one at a time faster then what we have currently?

    Parameters
    ----------
    cfg : object Configuration
        Configuration object containing all the configurable options in the OrcaNet scripts.
    files_dict : dict
        The name of every input as a key (can be multiple), the filepath of a h5py file to read samples from as values.
    f_size : int or None
        Specifies the filesize (#images) of the .h5 file if not the whole .h5 file
        should be used for yielding the xs/ys arrays. This is important if you run fit_generator(epochs>1) with
        a filesize (and hence # of steps) that is smaller than the .h5 file.
    zero_center_image : dict
        Mean image of the dataset used for zero-centering. Every input as a key, ndarray as values.
    yield_mc_info : bool
        Specifies if mc-infos (y_values) should be yielded as well.
        The mc-infos are used for evaluation after training and testing is finished.
    shuffle : bool
        Randomize the order in which batches are read from the file. Significantly reduces read out speed.

    Yields
    ------
    xs : dict
        Data for the model train on.
    ys : dict
        Labels for the model to train on.
    mc_info : ndarray, optional
        Mc info from the file. Only yielded if yield_mc_info is True.

    """
    batchsize = cfg.batchsize
    # name of the datagroups in the file
    samples_key = cfg.key_samples
    mc_key = cfg.key_labels

    # If the batchsize is larger than the f_size, make batchsize smaller or nothing would be yielded
    if f_size is not None:
        if f_size < batchsize:
            batchsize = f_size

    while 1:
        # a dict with the names of list inputs as keys, and the opened h5 files as values.
        files = {}
        file_lengths = []
        # open the files and make sure they have the same length
        for input_key in files_dict:
            files[input_key] = h5py.File(files_dict[input_key], 'r')
            file_lengths.append(len(files[input_key][samples_key]))
        if not file_lengths.count(file_lengths[0]) == len(file_lengths):
            raise AssertionError("All data files must have the same length! Yours have:\n " + str(file_lengths))

        if f_size is None:
            f_size = file_lengths[0]
        # number of full batches available
        total_no_of_batches = int(f_size/batchsize)
        # positions of the samples in the file
        sample_pos = np.arange(total_no_of_batches, step=batches_at_once) * batchsize
        if shuffle:
            np.random.shuffle(sample_pos)

        for sample_n in sample_pos:
            # Read one batch of samples from the files and zero center
            # A dict with every input name as key, and a batch of data as values
            xs = {}
            for input_key in files:
                xs[input_key] = files[input_key][samples_key][sample_n: sample_n + batches_at_once*batchsize]
                if zero_center_image is not None:
                    xs[input_key] = np.subtract(xs[input_key], zero_center_image[input_key])
            # Get labels for the nn. Since the labels are hopefully the same for all the files, use the ones from the first
            y_values = list(files.values())[0][mc_key][sample_n:sample_n + batches_at_once*batchsize]

            for bsample_no in np.arange(batches_at_once)*batchsize:
                bxs = {}
                for input_key in xs:
                    bxs[input_key] = xs[input_key][bsample_no: bsample_no + batchsize]
                by_values = y_values[bsample_no: bsample_no + batchsize]

                # Modify the samples and the labels batchwise
                if cfg.sample_modifier is not None:
                    bxs = cfg.sample_modifier(bxs)

                # if swap_col is not None:
                #     xs = get_input_images(xs, swap_col, str_ident)
                bys = by_values  # get_labels(by_values, class_type)

                if not yield_mc_info:
                    yield bxs, bys
                else:
                    yield bxs, bys, by_values

        # for i in range(n_files):
        #     files[i].close()


def test_generators(batches, intermediate_log, functions, func_kwargs):
    """
    Test the time of generators.

    Parameters
    ----------
    batches : int
        how many batches to read out
    intermediate_log : int
        show current speeds after this many batches
    functions : list
        functions to test
    func_kwargs : list
        kwargs for these functions

    """

    list_file = "/home/woody/capn/mppi013h/Code/OrcaNet/test/time_test_list.toml"
    zero_center_folder = "/home/woody/capn/mppi013h/Code/work/zero_center_folder/"
    modelfile = "/home/woody/capn/mppi013h/Code/OrcaNet/examples/settings_files/example_model.toml"

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'
    orca = Organizer("./test/", list_file)
    orca.cfg.zero_center_folder = zero_center_folder
    train_files = next(orca.io.yield_files("train"))
    # cfg.use_local_node()
    print(batches, "batches from file", train_files)
    print("Batchsize:", orca.cfg.batchsize)
    f = h5py.File(list(train_files.values())[0], "r")
    print(f["x"].shape)
    xs_mean = None  # np.ones(f["x"].shape[1:])  # load_zero_center_data(cfg)
    f.close()

    orca.cfg.import_model_file(modelfile)
    model = build_nn_model(orca)
    # model.summary()

    # No large dataset in the new ys format exists, so use dummy data to train on instead
    ys = np.ones((orca.cfg.batchsize, 16))
    dtypes = [('event_id', '<f8'), ('particle_type', '<f8'), ('energy', '<f8'), ('is_cc', '<f8'), ('bjorkeny', '<f8'),
              ('dir_x', '<f8'), ('dir_y', '<f8'), ('dir_z', '<f8'), ('time_interaction', '<f8'), ('run_id', '<f8'),
              ('vertex_pos_x', '<f8'), ('vertex_pos_y', '<f8'), ('vertex_pos_z', '<f8'),
              ('time_residual_vertex', '<f8'),
              ('prod_ident', '<f8'), ('group_id', '<i8')]
    ys = ys.ravel().view(dtype=dtypes)
    ys = get_labels(ys, [0, 'energy_dir_bjorken-y_vtx_errors'])

    print("\n")
    for f_no, func in enumerate(functions):
        print("------------- Function", f_no, " -------------")
        generator = func(orca.cfg, train_files, zero_center_image=xs_mean, **func_kwargs[f_no])
        print("Shape of batches:", list(next(generator)[0].values())[0].shape, "\n")
        average_read_time, average_model_time = [], []
        for i in range(batches):
            read_start = time.time()
            x = list(next(generator)[0].values())[0]
            average_read_time.append(time.time() - read_start)

            model_start = time.time()
            model.train_on_batch(x, ys)
            average_model_time.append((time.time() - model_start))

            if i % intermediate_log == 0 and i != 0:
                print("At", i, "batches:  ", end="")
                print("Last {} batches:\tRead: {:.4g}\tTrain: {:.4g}".format(intermediate_log, np.average(average_read_time[i-200:i]), np.average(average_model_time[i-200:i])))
                # print("Total:\tRead:", np.average(average_read_time), "predict:", np.average(average_model_time))

        print("\nAverage read time per batch:\t{:.4g}".format(np.average(average_read_time)))
        print("Average train time per batch:\t{:.4g}".format(np.average(average_model_time)))


if __name__ == "__main__":
    batches = 500
    intermediate_log = 100
    functions = (generate_batches_from_hdf5_file_tweak, generate_batches_from_hdf5_file_tweak)
    func_kwargs = [{"batches_at_once": 10, "shuffle": True}, {"batches_at_once": 10, }]

    test_generators(batches, intermediate_log, functions, func_kwargs)



from unittest import TestCase
from unittest.mock import patch
import os
import h5py
import numpy as np
from keras.models import Model
from keras.layers import Dense, Input, Concatenate, Flatten
import shutil

from orcanet.core import Configuration
from orcanet.in_out import HistoryHandler, IOHandler


class TestIOHandler(TestCase):
    @classmethod
    def setUpClass(cls):
        # super(TestIOHandler, cls).setUpClass()
        cls.temp_dir = os.path.join(os.path.dirname(__file__), ".temp",
                                    "test_in_out")
        os.mkdir(cls.temp_dir)
        cls.init_dir = os.getcwd()
        os.chdir(cls.temp_dir)
        # make some dummy data
        cls.n_bins = {'input_A': (2, 3), 'input_B': (2, 3)}
        cls.train_sizes = [30, 50]
        cls.val_sizes = [40, ]
        cls.file_names = (
            "/input_A_train_1.h5",
            "/input_A_train_2.h5",
            "/input_B_train_1.h5",
            "/input_B_train_2.h5",
            "/input_A_val_1.h5",
            "/input_B_val_1.h5",
        )
        cls.train_A_file_1 = {
            "path": cls.temp_dir + cls.file_names[0],
            "shape": cls.n_bins["input_A"],
            "value_xs": 1.1,
            "value_ys": 1.2,
            "size": cls.train_sizes[0],
        }
        cls.train_A_file_2 = {
            "path": cls.temp_dir + cls.file_names[1],
            "shape": cls.n_bins["input_A"],
            "value_xs": 1.3,
            "value_ys": 1.4,
            "size": cls.train_sizes[1],
        }
        cls.train_B_file_1 = {
            "path": cls.temp_dir + cls.file_names[2],
            "shape": cls.n_bins["input_B"],
            "value_xs": 2.1,
            "value_ys": 2.2,
            "size": cls.train_sizes[0],
        }
        cls.train_B_file_2 = {
            "path": cls.temp_dir + cls.file_names[3],
            "shape": cls.n_bins["input_B"],
            "value_xs": 2.3,
            "value_ys": 2.4,
            "size": cls.train_sizes[1],
        }
        cls.val_A_file_1 = {
            "path": cls.temp_dir + cls.file_names[4],
            "shape": cls.n_bins["input_A"],
            "value_xs": 3.1,
            "value_ys": 3.2,
            "size": cls.val_sizes[0],
        }
        cls.val_B_file_1 = {
            "path": cls.temp_dir + cls.file_names[5],
            "shape": cls.n_bins["input_B"],
            "value_xs": 4.1,
            "value_ys": 4.2,
            "size": cls.val_sizes[0],
        }
        cls.train_A_file_1_ctnt = save_dummy_h5py(**cls.train_A_file_1)
        cls.train_A_file_2_ctnt = save_dummy_h5py(**cls.train_A_file_2)
        cls.train_B_file_1_ctnt = save_dummy_h5py(**cls.train_B_file_1)
        cls.train_B_file_2_ctnt = save_dummy_h5py(**cls.train_B_file_2)
        cls.val_A_file_1_ctnt = save_dummy_h5py(**cls.val_A_file_1)
        cls.val_B_file_1_ctnt = save_dummy_h5py(**cls.val_B_file_1)

    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.output_folder = self.data_folder + "/dummy_model"

        list_file = self.data_folder + "/in_out_test_list.toml"
        config_file = None

        cfg = Configuration(self.output_folder, list_file, config_file)
        self.batchsize = 3
        cfg.batchsize = self.batchsize
        self.io = IOHandler(cfg)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.train_A_file_1["path"])
        os.remove(cls.train_A_file_2["path"])
        os.remove(cls.train_B_file_1["path"])
        os.remove(cls.train_B_file_2["path"])
        os.remove(cls.val_A_file_1["path"])
        os.remove(cls.val_B_file_1["path"])

        os.chdir(cls.init_dir)
        os.rmdir(cls.temp_dir)

    def test_use_local_node(self):
        temp_temp_dir = self.temp_dir + "/scratch"
        os.mkdir(temp_temp_dir)
        if "TMPDIR" in os.environ:
            tempdir_environ = os.environ["TMPDIR"]
        else:
            tempdir_environ = None
        try:
            os.environ["TMPDIR"] = temp_temp_dir
            scratch_dir = temp_temp_dir

            target_dirs_train = {
                 "input_A": (scratch_dir + self.file_names[0], scratch_dir + self.file_names[1]),
                 "input_B": (scratch_dir + self.file_names[2], scratch_dir + self.file_names[3]),
            }
            target_dirs_val = {
                 "input_A": (scratch_dir + self.file_names[4], ),
                 "input_B": (scratch_dir + self.file_names[5], ),
            }

            self.io.use_local_node()

            value = self.io.get_local_files("train")
            self.assertDictEqual(target_dirs_train, value)

            value = self.io.get_local_files("val")
            self.assertDictEqual(target_dirs_val, value)
        finally:
            if tempdir_environ is not None:
                os.environ["TMPDIR"] = tempdir_environ
            else:
                os.environ.pop("TMPDIR")

            shutil.rmtree(temp_temp_dir)

    def test_check_connections_no_sample(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        self.io.cfg.label_modifier = get_dummy_label_modifier(output_shapes.keys())
        model = build_dummy_model(input_shapes, output_shapes)

        self.io.check_connections(model)

    def test_check_connections_ok_sample(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        def sample_modifier(samples):
            return {'input_A': samples["input_A"], 'input_B': samples["input_B"]}

        self.io.cfg.label_modifier = get_dummy_label_modifier(
            output_shapes.keys())
        self.io.cfg.sample_modifier = sample_modifier

        model = build_dummy_model(input_shapes, output_shapes)
        self.io.check_connections(model)

    def test_check_connections_wrong_sample(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        def sample_modifier(samples):
            return {'input_A': samples["input_A"]}

        self.io.cfg.label_modifier = get_dummy_label_modifier(
            output_shapes.keys())
        self.io.cfg.sample_modifier = sample_modifier

        model = build_dummy_model(input_shapes, output_shapes)
        with self.assertRaises(ValueError):
            self.io.check_connections(model)

    def test_check_connections_wrong_label(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        self.io.cfg.label_modifier = get_dummy_label_modifier(["out_A"])
        model = build_dummy_model(input_shapes, output_shapes)

        with self.assertRaises(ValueError):
            self.io.check_connections(model)

    def test_check_connections_no_label(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        model = build_dummy_model(input_shapes, output_shapes)

        with self.assertRaises(ValueError):
            self.io.check_connections(model)

    def test_check_connections_auto_label(self):
        input_shapes = self.n_bins
        output_shapes = {
            "mc_A": 1,
            "mc_B": 1,
        }

        model = build_dummy_model(input_shapes, output_shapes)

        self.io.check_connections(model)

    def test_get_n_bins(self):
        value = self.io.get_n_bins()
        self.assertSequenceEqual(value, self.n_bins)

    def test_get_input_shape(self):
        value = self.io.get_input_shapes()
        self.assertSequenceEqual(value, self.n_bins)

        def sample_modifier(samples):
            return {'input_A': samples["input_A"], }
        self.io.cfg.sample_modifier = sample_modifier

        value = self.io.get_input_shapes()
        self.assertEqual(value, {"input_A": self.n_bins["input_A"]})

    def test_get_file_sizes_train(self):
        value = self.io.get_file_sizes("train")
        self.assertSequenceEqual(value, self.train_sizes)

    def test_get_batch_xs(self):
        value = self.io.get_batch()
        target = {
            "input_A": self.train_A_file_1_ctnt[0][:self.batchsize],
            "input_B": self.train_B_file_1_ctnt[0][:self.batchsize],
        }
        assert_dict_arrays_equal(value[0], target)

    def test_get_batch_ys(self):
        value = self.io.get_batch()
        target = {
            "input_A": self.train_A_file_1_ctnt[1][:self.batchsize],
            "input_B": self.train_B_file_1_ctnt[1][:self.batchsize],
        }
        assert_equal_struc_array(value[1], target["input_A"])

    def test_get_latest_epoch(self):
        value = self.io.get_latest_epoch()
        target = (2, 1)
        self.assertSequenceEqual(value, target)

    def test_get_latest_epoch_epoch_1(self):
        value = self.io.get_latest_epoch(epoch=1)
        target = (1, 2)
        self.assertSequenceEqual(value, target)

    def test_get_latest_epoch_no_files_but_epoch_given(self):
        with self.assertRaises(ValueError):
            self.io.get_latest_epoch(epoch=3)

    def test_get_next_epoch_none(self):
        value = self.io.get_next_epoch(None)
        target = (1, 1)
        self.assertSequenceEqual(value, target)

    def test_get_next_epoch_1_1(self):
        value = self.io.get_next_epoch((1, 1))
        target = (1, 2)
        self.assertSequenceEqual(value, target)

    def test_get_next_epoch_1_2(self):
        value = self.io.get_next_epoch((1, 2))
        target = (2, 1)
        self.assertSequenceEqual(value, target)

    def test_get_previous_epoch_2_1(self):
        value = self.io.get_previous_epoch((2, 1))
        target = (1, 2)
        self.assertSequenceEqual(value, target)

    def test_get_previous_epoch_1_2(self):
        value = self.io.get_previous_epoch((1,2))
        target = (1, 1)
        self.assertSequenceEqual(value, target)

    def test_get_model_path(self):
        value = self.io.get_model_path(1, 1)
        target = self.output_folder + '/saved_models/model_epoch_1_file_1.h5'
        self.assertEqual(value, target)

    def test_get_model_path_latest(self):
        value = self.io.get_model_path(-1, -1)
        target = self.output_folder + '/saved_models/model_epoch_2_file_1.h5'
        self.assertEqual(value, target)

    def test_get_pred_path(self):
        value = self.io.get_pred_path(2, 1)
        target = self.output_folder + '/predictions/pred_model_epoch_2_file_1' \
                                      '_on_in_out_test_list_val_files.h5'
        self.assertEqual(value, target)

    def test_get_local_files_train(self):
        value = self.io.get_local_files("train")
        target = {
            'input_A': ('input_A_train_1.h5', 'input_A_train_2.h5'),
            'input_B': ('input_B_train_1.h5', 'input_B_train_2.h5'),
        }
        self.assertDictEqual(value, target)

    def test_get_local_files_val(self):
        value = self.io.get_local_files("val")
        target = {
            'input_A': ('input_A_val_1.h5',),
            'input_B': ('input_B_val_1.h5',)
        }
        self.assertDictEqual(value, target)

    def test_get_no_of_files_train(self):
        value = self.io.get_no_of_files("train")
        target = 2
        self.assertEqual(value, target)

    def test_get_no_of_files_val(self):
        value = self.io.get_no_of_files("val")
        target = 1
        self.assertEqual(value, target)

    def test_yield_files_train(self):
        file_paths = self.io.yield_files("train")
        target = (
            {
                'input_A': 'input_A_train_1.h5',
                'input_B': 'input_B_train_1.h5',
            },
            {
                'input_A': 'input_A_train_2.h5',
                'input_B': 'input_B_train_2.h5',
            },
        )
        for i, value in enumerate(file_paths):
            self.assertDictEqual(value, target[i])

    def test_yield_files_val(self):
        file_paths = self.io.yield_files("val")
        target = (
            {
                'input_A': 'input_A_val_1.h5',
                'input_B': 'input_B_val_1.h5',
            },
        )
        for i, value in enumerate(file_paths):
            self.assertDictEqual(value, target[i])

    def test_get_file_train(self):
        value = self.io.get_file("train", 2)
        target = {
                'input_A': 'input_A_train_2.h5',
                'input_B': 'input_B_train_2.h5',
        }
        self.assertDictEqual(value, target)

    def test_get_file_val(self):
        value = self.io.get_file("val", 1)
        target = {
                'input_A': 'input_A_val_1.h5',
                'input_B': 'input_B_val_1.h5',
        }
        self.assertDictEqual(value, target)


class TestHistoryHandler(TestCase):
    def setUp(self):
        self.output_folder = os.path.join(os.path.dirname(__file__),
                                          "data", "dummy_model")
        self.summary_filename = os.path.join(self.output_folder, "summary.txt")
        self.summary_filename_2 = os.path.join(self.output_folder, "summary_2.txt")
        self.train_log_folder = os.path.join(self.output_folder, "train_log")

        self.history = HistoryHandler(self.summary_filename,
                                      self.train_log_folder)

    def test_get_metrics(self):
        metrics = self.history.get_metrics()
        target = ["loss", "acc"]
        np.testing.assert_array_equal(metrics, target)

    def test_get_summary_data(self):
        summary_data = self.history.get_summary_data()
        target = np.array(
            [(0.03488, 0.005, 0.07776, 0.05562, 0.971, 0.9808),
             (0.06971, 0.00465, 0.05034, np.nan, 0.9822, np.nan)],
            dtype=[('Epoch', '<f8'), ('LR', '<f8'), ('train_loss', '<f8'),
                   ('val_loss', '<f8'), ('train_acc', '<f8'), ('val_acc', '<f8')])
        assert_equal_struc_array(summary_data, target)

    def test_get_train_data(self):
        train_data = self.history.get_train_data()
        print(train_data)
        target = np.array(
            [(250., 0.00018755, 0.487258, 0.756687),
             (500., 0.00056265, 0.291661, 0.874188),
             (250., 0.0350691, 0.0589843, 0.977938),
             (500., 0.0354442, 0.0554589, 0.980375)],
            dtype=[('Batch', '<f8'), ('Batch_float', '<f8'),
                   ('loss', '<f8'), ('acc', '<f8')])
        assert_equal_struc_array(train_data, target)

    def test_get_column_names(self):
        column_names = self.history.get_column_names()
        target = ('Epoch', 'LR', 'train_loss', 'val_loss', 'train_acc', 'val_acc')
        self.assertSequenceEqual(column_names, target)

    def test_get_state(self):
        self.history = HistoryHandler(self.summary_filename_2,
                                      self.train_log_folder)
        state = self.history.get_state()
        print(state)
        target = [
            {'epoch': 0.1, 'is_trained': True, 'is_validated': True},
            {'epoch': 0.2, 'is_trained': False, 'is_validated': True},
            {'epoch': 0.3, 'is_trained': True, 'is_validated': False},
            {'epoch': 0.4, 'is_trained': False, 'is_validated': False},
        ]
        self.assertSequenceEqual(state, target)

    def test_plot_metric_onknown_metric(self):
        with self.assertRaises(ValueError):
            self.history.plot_metric("test")

    @patch('orcanet.in_out.plot_history')
    def test_plot_metric_loss(self, mock_plot_history):
        def plot_history(train_data, val_data, **kwargs):
            return train_data, val_data, kwargs
        mock_plot_history.side_effect = plot_history

        value_train, value_val, value_kwargs = self.history.plot_metric("loss")

        target_train = [
            np.array([0.000187551, 0.000562653, 0.0350691, 0.0354442]),
            np.array([0.487258, 0.291661, 0.0589843, 0.0554589]),
        ]

        target_val = [
            np.array([0.03488, 0.06971]),
            np.array([0.05562, np.nan]),
        ]
        target_kwargs = {'y_label': 'loss'}

        np.testing.assert_array_almost_equal(target_train, value_train)
        np.testing.assert_array_almost_equal(target_val, value_val)
        self.assertDictEqual(target_kwargs, value_kwargs)

    @patch('orcanet.in_out.plot_history')
    def test_plot_metric_acc(self, mock_plot_history):
        def plot_history(train_data, val_data, **kwargs):
            return train_data, val_data, kwargs
        mock_plot_history.side_effect = plot_history

        value_train, value_val, value_kwargs = self.history.plot_metric("acc")

        target_train = [
            np.array([0.000187551, 0.000562653, 0.0350691, 0.0354442]),
            np.array([0.756687, 0.874188, 0.977938, 0.980375]),
        ]

        target_val = [
            np.array([0.03488, 0.06971]),
            np.array([0.9808, np.nan]),
        ]
        target_kwargs = {'y_label': 'acc'}

        np.testing.assert_array_almost_equal(target_train, value_train)
        np.testing.assert_array_almost_equal(target_val, value_val)
        self.assertDictEqual(target_kwargs, value_kwargs)

    @patch('orcanet.in_out.plot_history')
    def test_plot_lr(self, mock_plot_history):
        def plot_history(train_data, val_data, **kwargs):
            return train_data, val_data, kwargs
        mock_plot_history.side_effect = plot_history

        value_train, value_val, value_kwargs = self.history.plot_lr()

        target_train = None

        target_val = [
            np.array([0.03488, 0.06971]),
            np.array([0.005, 0.00465])
        ]
        target_kwargs = {
            'y_label': 'Learning rate',
            'legend': False
        }

        self.assertEqual(target_train, value_train)
        np.testing.assert_array_almost_equal(target_val, value_val)
        self.assertDictEqual(target_kwargs, value_kwargs)


def assert_equal_struc_array(a, b):
    """  np.testing.assert_array_equal does not work for arrays containing nans...
     so test individual instead. """
    np.testing.assert_array_equal(a.dtype, b.dtype)
    for name in a.dtype.names:
        np.testing.assert_almost_equal(a[name], b[name])


def save_dummy_h5py(path, shape, value_xs=1., value_ys=1., size=50):
    xs = np.ones((size,) + shape) * value_xs

    dtypes = [('mc_A', '<f8'), ('mc_B', '<f8'), ]
    ys = (np.ones((size, 2))*value_ys).ravel().view(dtype=dtypes)

    with h5py.File(path, 'w') as h5f:
        h5f.create_dataset('x', data=xs, dtype='<f8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)

    return xs, ys


def build_dummy_model(input_shapes, output_shapes):
    """

    Parameters
    ----------
    input_shapes : dict
    output_shapes : dict

    Returns
    -------
    model : keras model

    """
    inputs = {}
    for name, shape in input_shapes.items():
        inputs[name] = Input(shape, name=name)
    conc = Concatenate()(list(inputs.values()))
    flat = Flatten()(conc)

    outputs = {}
    for name, shape in output_shapes.items():
        outputs[name] = Dense(shape, name=name)(flat)

    model = Model(list(inputs.values()), list(outputs.values()))
    return model


def get_dummy_label_modifier(output_names):
    def label_modifier(mc_info):
        particle = mc_info["mc_A"]
        y_true = dict()
        for output_name in output_names:
            y_true[output_name] = particle
        return y_true
    return label_modifier


def assert_dict_arrays_equal(a, b):
    for key, value in a.items():
        assert key in b
        np.testing.assert_array_almost_equal(a[key], b[key])




#!/usr/bin/env python
from setuptools import setup

with open('requirements.txt') as fobj:
    requirements = [l.strip() for l in fobj.readlines()]

setup(
    name='orcanet',
    description='Runs Neural Networks for usage in the KM3NeT project',
    url='https://git.km3net.de/ml/OrcaNet',
    author='Michael Moser, Stefan Reck',
    author_email='mmoser@km3net.de, michael.m.moser@fau.de, stefan.reck@fau.de',
    license='AGPL',
    install_requires=requirements,
    packages=[
        "orcanet", "orcanet_contrib"
    ],
    include_package_data=True,

    setup_requires=['setuptools_scm'],
    use_scm_version={
        'write_to': 'orcanet/version.txt',
        'tag_regex': r'^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$',
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
    ],

    entry_points={'console_scripts': ['orcatrain=orcanet_contrib.parser_orcatrain:main',
                                      'orcapred=orcanet_contrib.parser_orcapred:main',
                                      ]},

)

__author__ = 'Michael Moser and Stefan Reck'
